import dataclasses
import datetime
import json
from typing import Any, cast

from pydantic import BaseModel

from harbory.common.base58 import b58encode
from harbory.common.hashutils import hash_object_bytes
from harbory.common.typeutils import is_namedtuple
from harbory.types import JsonValue


def object_fingerprint(obj: Any) -> str:
    return b58encode(hash_object_bytes(obj)).decode()


def as_jsonvalue(value: Any) -> JsonValue:
    return cast(JsonValue, json.loads(json.dumps(value, cls=WorkflowJSONEncoder)))


class WorkflowJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        from .colt import WorkflowRef
        from .graph import WorkflowGraph
        from .step import WorkflowStepInfo

        if isinstance(obj, (WorkflowGraph, WorkflowStepInfo)):
            return obj.to_dict()
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if is_namedtuple(obj):
            return obj._asdict()
        if isinstance(obj, (set, frozenset, tuple)):
            return list(obj)
        if isinstance(obj, WorkflowRef):
            return obj.config
        if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
            return dataclasses.asdict(obj)
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        return super().default(obj)



from typing import Generic, Iterator

from colt import Registrable

from .types import T_Example


class DataSource(Generic[T_Example], Registrable):
    def load(self) -> Iterator[T_Example]:
        raise NotImplementedError


class DataSink(Generic[T_Example], Registrable):
    def save(self, data: Iterator[T_Example]) -> None:
        raise NotImplementedError



from __future__ import annotations

import ast
import itertools
import textwrap


class ASTNormalizer(ast.NodeTransformer):
    def __init__(self) -> None:
        self.global_name_counter = itertools.count()
        self.scoped_name_counter = itertools.count()
        self.global_name_mapping: dict[str, str] = {}
        self.scope_stack: list[dict[str, str]] = [{}]

    def find_name_in_scopes(self, name: str) -> str | None:
        for scope in reversed(self.scope_stack):
            if name in scope:
                return scope[name]
        return None

    def visit_Name(self, node: ast.Name) -> ast.Name:
        new_name = self.find_name_in_scopes(node.id)
        if new_name is not None:
            node.id = new_name
        return node

    def visit_arg(self, node: ast.arg) -> ast.arg:
        new_name = self.find_name_in_scopes(node.arg)
        if new_name is not None:
            node.arg = new_name
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        if node.name not in self.global_name_mapping:
            self.global_name_mapping[node.name] = f"var{next(self.global_name_counter)}"
        node.name = self.global_name_mapping[node.name]
        self.scope_stack.append({})
        self.scoped_name_counter = itertools.count()
        for arg in node.args.args:
            new_name = f"var{next(self.scoped_name_counter)}"
            self.scope_stack[-1][arg.arg] = new_name
        self.generic_visit(node)
        self.scope_stack.pop()
        return node

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        for target in node.targets:
            if isinstance(target, ast.Name):
                new_name = self.find_name_in_scopes(target.id)
                if new_name is None:
                    new_name = f"var{next(self.scoped_name_counter)}"
                    self.scope_stack[-1][target.id] = new_name
                target.id = new_name
        self.generic_visit(node)
        return node


def normalize_source(source: str) -> str:
    source = textwrap.dedent(source)
    tree = ast.parse(source)
    normalizer = ASTNormalizer()
    tree = normalizer.visit(tree)
    return ast.dump(tree)



import json
import shutil
import uuid
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from typing import ClassVar, Optional, TypeVar, Union

from colt import Registrable
from filelock import FileLock

from harbory.common.logutils import LogCapture

from .cache import FilesystemWorkflowCache, MemoryWorkflowCache, WorkflowCache
from .callback import MultiWorkflowCallback, WorkflowCallback
from .colt import COLT_BUILDER
from .constants import WORKFLOW_DEFAULT_DIRECTORY
from .executor import WorkflowExecutionContext, WorkflowExecutionID, WorkflowExecutionInfo, WorkflowExecutor
from .graph import WorkflowGraph
from .step import WorkflowStepContext, WorkflowStepInfo, get_step_logger_from_info
from .utils import WorkflowJSONEncoder

T_WorkflowOrganizer = TypeVar("T_WorkflowOrganizer", bound="WorkflowOrganizer")


class WorkflowOrganizer(Registrable):
    def __init__(
        self,
        cache: "WorkflowCache",
        callbacks: Optional[Union[WorkflowCallback, Sequence[WorkflowCallback]]],
    ) -> None:
        if isinstance(callbacks, WorkflowCallback):
            callbacks = [callbacks]

        self.cache = cache
        self.callback = MultiWorkflowCallback(callbacks or [])

    def run(
        self,
        executor: "WorkflowExecutor",
        execution: Union[WorkflowGraph, WorkflowExecutionInfo],
    ) -> WorkflowExecutionInfo:
        with executor:
            return executor(
                execution,
                cache=self.cache,
                callback=self.callback,
            )

    def get(self, execution_id: WorkflowExecutionID) -> Optional[WorkflowExecutionInfo]:
        return None

    def exists(self, execution_id: WorkflowExecutionID) -> bool:
        return self.get(execution_id) is not None

    def remove(self, execution_id: WorkflowExecutionID) -> None:
        pass


@WorkflowOrganizer.register("memory")
class MemoryWorkflowOrganizer(WorkflowOrganizer):
    def __init__(
        self,
        callbacks: Optional[Union[WorkflowCallback, Sequence[WorkflowCallback]]] = None,
    ) -> None:
        super().__init__(MemoryWorkflowCache(), callbacks)


@WorkflowOrganizer.register("filesystem")
class FilesystemWorkflowOrganizer(WorkflowOrganizer):
    _CACHE_DIRNAME: ClassVar[str] = "cache"
    _EXECUTIONS_DIRNAME: ClassVar[str] = "executions"
    _DEFAULT_DIRECTORY: ClassVar[Path] = WORKFLOW_DEFAULT_DIRECTORY

    class _Callback(WorkflowCallback):
        _LOG_FILENAME: ClassVar[str] = "out.log"
        _STEPS_DIRNAME: ClassVar[str] = "steps"
        _LOCK_FILENAME: ClassVar[str] = "__lock__"
        _STEP_FILENAME: ClassVar[str] = "step.json"
        _STATE_FILENAME: ClassVar[str] = "state.json"
        _RESULT_FILENAME: ClassVar[str] = "result"
        _EXECUTION_FILENAME: ClassVar[str] = "execution.json"

        def __init__(self, organizer: "FilesystemWorkflowOrganizer") -> None:
            self._organizer = organizer
            self._execution_log: Optional[LogCapture] = None
            self._step_log: dict[WorkflowStepInfo, LogCapture] = {}

        def _generate_new_execution_id(self) -> WorkflowExecutionID:
            execution_id = uuid.uuid4().hex[:8]
            while (self._organizer.executions_directory / execution_id).exists():
                execution_id = uuid.uuid4().hex[:8]
            return WorkflowExecutionID(execution_id)

        def _get_execution_directory(self, execution_info: "WorkflowExecutionInfo") -> Path:
            if execution_info.id is None:
                raise ValueError("Execution ID is not set")
            return self._organizer.executions_directory / execution_info.id

        def _get_step_directory(
            self,
            execution_info: "WorkflowExecutionInfo",
            step_info: "WorkflowStepInfo",
        ) -> Path:
            execution_directory = self._get_execution_directory(execution_info)
            return execution_directory / self._STEPS_DIRNAME / step_info.name

        def on_execution_start(
            self,
            execution_context: "WorkflowExecutionContext",
        ) -> None:
            execution_info = execution_context.info
            if execution_info.id is None:
                execution_info.id = self._generate_new_execution_id()

            execution_directory = self._get_execution_directory(execution_info)
            execution_directory.mkdir(parents=True, exist_ok=True)

            with FileLock(execution_directory / self._LOCK_FILENAME):
                self._execution_log = LogCapture((execution_directory / self._LOG_FILENAME).open("w"))
                self._execution_log.start()

                with open(execution_directory / self._EXECUTION_FILENAME, "w") as jsonfile:
                    json.dump(
                        execution_info,
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )
                with open(execution_directory / self._STATE_FILENAME, "w") as jsonfile:
                    json.dump(
                        execution_context.state,
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )

        def on_execution_end(
            self,
            execution_context: "WorkflowExecutionContext",
        ) -> None:
            execution_info = execution_context.info
            execution_directory = self._get_execution_directory(execution_info)

            with FileLock(execution_directory / self._LOCK_FILENAME):
                if self._execution_log is not None:
                    self._execution_log.stop()
                    self._execution_log.stream.close()
                    self._execution_log = None
                with open(execution_directory / self._STATE_FILENAME, "w") as jsonfile:
                    json.dump(
                        execution_context.state,
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )

        def on_step_start(
            self,
            step_context: "WorkflowStepContext",
            execution_context: "WorkflowExecutionContext",
        ) -> None:
            step_info = step_context.info
            execution_info = execution_context.info

            step_directory = self._get_step_directory(execution_info, step_info)
            step_directory.mkdir(parents=True, exist_ok=True)

            with FileLock(step_directory / self._LOCK_FILENAME):
                self._step_log[step_info] = LogCapture(
                    (step_directory / self._LOG_FILENAME).open("w"),
                    logger=get_step_logger_from_info(step_info),
                )
                self._step_log[step_info].start()

                with open(step_directory / self._STEP_FILENAME, "w") as jsonfile:
                    json.dump(
                        step_info.to_dict(),
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )
                with open(step_directory / self._STATE_FILENAME, "w") as jsonfile:
                    json.dump(
                        step_context.state,
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )

        def on_step_end(
            self,
            step_context: "WorkflowStepContext",
            execution_context: "WorkflowExecutionContext",
        ) -> None:
            step_info = step_context.info
            execution_info = execution_context.info
            cache = execution_context.cache

            step_directory = self._get_step_directory(execution_info, step_info)
            cache_directory = self._organizer.cache_directory / step_info.fingerprint
            result_path = step_directory / self._RESULT_FILENAME
            with FileLock(step_directory / self._LOCK_FILENAME):
                if (step_log := self._step_log.pop(step_info, None)) is not None:
                    step_log.stop()
                    step_log.stream.close()
                with open(step_directory / self._STATE_FILENAME, "w") as jsonfile:
                    json.dump(
                        step_context.state,
                        jsonfile,
                        cls=WorkflowJSONEncoder,
                        indent=2,
                        ensure_ascii=False,
                    )
                if step_info in cache and not result_path.exists():
                    result_path.symlink_to(cache_directory)

    def __init__(
        self,
        directory: Optional[Union[str, PathLike]] = None,
        callbacks: Optional[Union[WorkflowCallback, Sequence[WorkflowCallback]]] = None,
    ) -> None:
        self._directory = Path(directory or self._DEFAULT_DIRECTORY).expanduser().resolve().absolute()

        if isinstance(callbacks, WorkflowCallback):
            callbacks = [callbacks]
        callbacks = [self._Callback(self)] + list(callbacks or [])

        super().__init__(FilesystemWorkflowCache(self.cache_directory), callbacks)

    @property
    def directory(self) -> Path:
        return self._directory

    @property
    def cache_directory(self) -> Path:
        return self._directory / self._CACHE_DIRNAME

    @property
    def executions_directory(self) -> Path:
        return self._directory / self._EXECUTIONS_DIRNAME

    def get(self, execution_id: WorkflowExecutionID) -> Optional[WorkflowExecutionInfo]:
        execution_directory = self.executions_directory / execution_id
        if not execution_directory.exists():
            return None

        execution_path = execution_directory / self._Callback._EXECUTION_FILENAME
        if not execution_path.exists():
            return None

        with open(execution_path, "r") as jsonfile:
            execution_info = COLT_BUILDER(json.load(jsonfile), WorkflowExecutionInfo)

        return execution_info

    def exists(self, execution_id: WorkflowExecutionID) -> bool:
        execution_directory = self.executions_directory / execution_id
        return execution_directory.exists()

    def remove(self, execution_id: WorkflowExecutionID) -> None:
        execution_directory = self.executions_directory / execution_id
        if not execution_directory.exists():
            return

        shutil.rmtree(execution_directory)



from typing import Final

from harbory.constants import DEFAULT_MLFACTORY_DIRECTORY, DEFAULT_WORKING_DIRECTORY

WORKFLOW_REFKEY: Final = "ref"
WORKFLOW_REFTYPE: Final = "ref"
WORKFLOW_DEFAULT_DIRECTORY: Final = DEFAULT_MLFACTORY_DIRECTORY / "workflow"
WORKFLOW_DEFAULT_SETTINGS_PATH: Final = DEFAULT_WORKING_DIRECTORY / "harbory.yml"
WORKFLOW_INTEGRATION_DIRECTORY: Final = DEFAULT_MLFACTORY_DIRECTORY / "integrations"



import dataclasses
import tarfile
import tempfile
from os import PathLike
from pathlib import Path
from typing import ClassVar, Generic, TypeVar, Union, cast

from .model import Model, T_Model

T_MachineryArchive = TypeVar("T_MachineryArchive", bound="MachineryArchive")


@dataclasses.dataclass
class MachineryArchive(Generic[T_Model]):
    model: T_Model

    _MODEL_FILENAME: ClassVar[str] = "model"

    def save(self, path: Union[str, PathLike]) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)

            model_path = working_dir / self._MODEL_FILENAME
            self.model.save(model_path)

            with tarfile.open(path, "w:gz") as tar:
                tar.add(model_path, arcname=self._MODEL_FILENAME)

    @classmethod
    def load(cls: type[T_MachineryArchive], path: Union[str, PathLike]) -> T_MachineryArchive:
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)

            with tarfile.open(path, "r:gz") as tar:
                tar.extractall(working_dir)

            model_path = working_dir / cls._MODEL_FILENAME
            model = cast(T_Model, Model.load(model_path))

        return cls(model)



import contextvars
import dataclasses
import datetime
from collections.abc import Mapping, Sequence
from enum import Enum
from importlib.metadata import version
from logging import getLogger
from types import TracebackType
from typing import Any, NewType, Optional, TypeVar, Union

from colt import Registrable

from harbory.common.git import GitInfo, get_git_info
from harbory.common.pkgutils import PackageInfo, get_installed_packages

from .cache import EmptyWorkflowCache, WorkflowCache
from .callback import EmptyWorkflowCallback, WorkflowCallback
from .graph import WorkflowGraph, WorkflowStepInfo
from .step import WorkflowStep, WorkflowStepContext, WorkflowStepState, WorkflowStepStatus

logger = getLogger(__name__)

T = TypeVar("T")
T_WorkflowExecutor = TypeVar("T_WorkflowExecutor", bound="WorkflowExecutor")

_EXECUTION_CONTEXT = contextvars.ContextVar[Optional["WorkflowExecutionContext"]]("_EXECUTION_CONTEXT", default=None)


WorkflowExecutionID = NewType("WorkflowExecutionID", str)


class WorkflowExecutionStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    FAILURE = "failure"
    CANCELED = "canceled"
    COMPLETED = "completed"


@dataclasses.dataclass
class WorkflowExecutionState:
    execution_id: Optional[WorkflowExecutionID] = None
    status: WorkflowExecutionStatus = WorkflowExecutionStatus.PENDING
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class WorkflowExecutionMetadata:
    version: str = version("harbory")
    git: Optional[GitInfo] = dataclasses.field(default_factory=get_git_info)
    environment: Mapping[str, str] = dataclasses.field(default_factory=dict)
    required_modules: Sequence[str] = dataclasses.field(default_factory=list)
    dependent_packages: Sequence[PackageInfo] = dataclasses.field(default_factory=get_installed_packages)


@dataclasses.dataclass
class WorkflowExecutionInfo:
    graph: WorkflowGraph

    id: Optional[WorkflowExecutionID] = None
    metadata: WorkflowExecutionMetadata = dataclasses.field(default_factory=WorkflowExecutionMetadata)


@dataclasses.dataclass
class WorkflowExecutionContext:
    info: WorkflowExecutionInfo
    state: WorkflowExecutionState
    cache: WorkflowCache = dataclasses.field(default_factory=EmptyWorkflowCache)
    callback: WorkflowCallback = dataclasses.field(default_factory=EmptyWorkflowCallback)


class WorkflowExecutor(Registrable):
    def __call__(
        self,
        graph_or_exection: Union[WorkflowGraph, WorkflowExecutionInfo],
        *,
        cache: Optional[WorkflowCache] = None,
        callback: Optional[WorkflowCallback] = None,
    ) -> WorkflowExecutionInfo:
        raise NotImplementedError

    def __enter__(self: T_WorkflowExecutor) -> T_WorkflowExecutor:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        pass


@WorkflowExecutor.register("default")
class DefaultWorkflowExecutor(WorkflowExecutor):
    def __call__(
        self,
        graph_or_exection: Union[WorkflowGraph, WorkflowExecutionInfo],
        *,
        cache: Optional[WorkflowCache] = None,
        callback: Optional[WorkflowCallback] = None,
    ) -> WorkflowExecutionInfo:
        cache = cache if cache is not None else EmptyWorkflowCache()
        callback = callback if callback is not None else EmptyWorkflowCallback()
        execution_info = (
            graph_or_exection
            if isinstance(graph_or_exection, WorkflowExecutionInfo)
            else WorkflowExecutionInfo(graph_or_exection)
        )

        execution_state = WorkflowExecutionState(
            execution_id=execution_info.id,
            status=WorkflowExecutionStatus.RUNNING,
            started_at=datetime.datetime.now(),
        )
        execution_context = WorkflowExecutionContext(execution_info, execution_state, cache, callback)

        callback.on_execution_start(execution_context)

        # NOTE: execution info can be updated by the callback
        execution_state = dataclasses.replace(execution_state, execution_id=execution_info.id)
        execution_context = dataclasses.replace(execution_context, state=execution_state)

        def _run_step(step_info: WorkflowStepInfo[WorkflowStep[T]]) -> T:
            assert cache is not None
            assert callback is not None

            step_state = WorkflowStepState(
                fingerprint=step_info.fingerprint,
                status=WorkflowStepStatus.RUNNING,
                started_at=datetime.datetime.now(),
            )

            step_context = WorkflowStepContext(step_info, step_state)

            if step_info in cache:
                logger.info(f"Cached value found for step {step_info.name}")
                result = cache[step_info]
            else:
                try:
                    callback.on_step_start(step_context, execution_context)
                    dependencies: Mapping[Union[int, str, Sequence[Union[int, str]]], Any] = {
                        path: _run_step(dep) for path, dep in step_info.dependencies
                    }
                    if set(dependencies.keys()) != set(path for path, _ in step_info.dependencies):
                        raise ValueError("Dependencies are not consistent with the graph")

                    step = step_info.step.construct(dependencies)
                    result = step(step_context)

                    if step_info.should_be_cached:
                        cache[step_info] = result
                        result = cache[step_info]
                except KeyboardInterrupt:
                    step_state = dataclasses.replace(step_state, status=WorkflowStepStatus.CANCELED)
                    step_context = dataclasses.replace(step_context, state=step_state)
                    raise
                except Exception as e:
                    step_state = dataclasses.replace(step_state, status=WorkflowStepStatus.FAILURE)
                    step_context = dataclasses.replace(step_context, state=step_state)
                    raise e
                else:
                    step_state = dataclasses.replace(step_state, status=WorkflowStepStatus.COMPLETED)
                    step_context = dataclasses.replace(step_context, state=step_state)
                finally:
                    step_state = dataclasses.replace(step_state, finished_at=datetime.datetime.now())
                    step_context = dataclasses.replace(step_context, state=step_state)
                    callback.on_step_end(step_context, execution_context)

            return result

        try:
            ctx = contextvars.copy_context()

            def execute() -> None:
                logger.info("Starting workflow execution %s...", execution_info.id)
                _EXECUTION_CONTEXT.set(execution_context)
                for step_info in execution_info.graph:
                    logger.info("Running step %s...", step_info.name)
                    _run_step(step_info)

            ctx.run(execute)

        except KeyboardInterrupt:
            execution_state = dataclasses.replace(execution_state, status=WorkflowExecutionStatus.CANCELED)
            execution_context = dataclasses.replace(execution_context, state=execution_state)
            raise
        except Exception as e:
            execution_state = dataclasses.replace(execution_state, status=WorkflowExecutionStatus.FAILURE)
            execution_context = dataclasses.replace(execution_context, state=execution_state)
            raise e
        else:
            execution_state = dataclasses.replace(execution_state, status=WorkflowExecutionStatus.COMPLETED)
            execution_context = dataclasses.replace(execution_context, state=execution_state)
        finally:
            execution_state = dataclasses.replace(execution_state, finished_at=datetime.datetime.now())
            callback.on_execution_end(execution_context)

        return execution_info


def use_execution_context() -> Optional[WorkflowExecutionContext]:
    return _EXECUTION_CONTEXT.get()



import tempfile
import uuid
from collections.abc import Iterator, Mapping
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import mlflow
from mlflow.entities import Experiment as MlflowExperiment
from mlflow.entities import Run as MlflowRun
from mlflow.tracking import artifact_utils
from mlflow.tracking.client import MlflowClient
from mlflow.tracking.context import registry as context_registry
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_RUN_NAME, MLFLOW_RUN_NOTE

from harbory.types import JsonValue
from harbory.workflow import (
    WorkflowExecutionID,
    WorkflowExecutionInfo,
    WorkflowExecutionState,
    WorkflowExecutionStatus,
    WorkflowStepInfo,
    WorkflowStepState,
    WorkflowStepStatus,
)
from harbory.workflow.utils import as_jsonvalue

MlflowParamValue = Union[int, float, str, None]
MlflowParams = dict[str, MlflowParamValue]


class WorkflowRunType(str, Enum):
    STEP = "step"
    EXECUTION = "execution"


class WorkflowCacheStatus(str, Enum):
    PENDING = "pending"
    ACTIVE = "active"
    INACTIVE = "inactive"


class MlflowRunStatus(str, Enum):
    RUNNING = "RUNNING"
    SCHEDULED = "SCHEDULED"
    FINISHED = "FINISHED"
    FAILED = "FAILED"
    KILLED = "KILLED"

    @classmethod
    def from_step_sttus(cls, status: WorkflowStepStatus) -> "MlflowRunStatus":
        if status == WorkflowStepStatus.PENDING:
            return cls.SCHEDULED
        if status == WorkflowStepStatus.RUNNING:
            return cls.RUNNING
        if status == WorkflowStepStatus.COMPLETED:
            return cls.FINISHED
        if status == WorkflowStepStatus.FAILURE:
            return cls.FAILED
        if status == WorkflowStepStatus.CANCELED:
            return cls.KILLED
        raise ValueError(f"Invalid step status: {status}")

    @classmethod
    def from_execution_status(cls, status: WorkflowExecutionStatus) -> "MlflowRunStatus":
        if status == WorkflowExecutionStatus.PENDING:
            return cls.SCHEDULED
        if status == WorkflowExecutionStatus.RUNNING:
            return cls.RUNNING
        if status == WorkflowExecutionStatus.COMPLETED:
            return cls.FINISHED
        if status == WorkflowExecutionStatus.FAILURE:
            return cls.FAILED
        if status == WorkflowExecutionStatus.CANCELED:
            return cls.KILLED
        raise ValueError(f"Invalid execution status: {status}")


class MlflowTag(str, Enum):
    # MLflow tags
    MLFLOW_PARENT_RUN_ID = MLFLOW_PARENT_RUN_ID
    MLFLOW_RUN_NAME = MLFLOW_RUN_NAME
    MLFLOW_RUN_NOTE = MLFLOW_RUN_NOTE

    # Custom tags
    MLFACTORY_RUN_TYPE = "harbory.workflow.run_type"
    MLFACTORY_STEP_FINGERPRINT = "harbory.workflow.step.fingerprint"
    MLFACTORY_STEP_CACHE_STATUS = "harbory.workflow.step.cache_status"


def flatten_params(d: Mapping[str, JsonValue]) -> MlflowParams:
    result = {}
    for key, value in d.items():
        if isinstance(value, (list, tuple)):
            value = {str(i): v for i, v in enumerate(value)}
        if isinstance(value, dict):
            result.update({f"{key}.{k}": v for k, v in flatten_params(value).items()})
        else:
            result[key] = value
    return result


def build_filter_string(
    run_type: Optional[WorkflowRunType] = None,
    step_info: Optional[Union[str, WorkflowStepInfo]] = None,
    execution_info: Optional[Union[str, WorkflowExecutionInfo]] = None,
    additional_filters: Optional[str] = None,
) -> str:
    conditions: list[str] = []
    if run_type is not None:
        conditions.append(f"tags.{MlflowTag.MLFACTORY_RUN_TYPE.value} = '{run_type.value}'")
    if step_info is not None:
        step_fingerprint = step_info.fingerprint if isinstance(step_info, WorkflowStepInfo) else step_info
        conditions.append(f"tags.{MlflowTag.MLFACTORY_STEP_FINGERPRINT.value} = '{step_fingerprint}'")
    if execution_info is not None:
        execution_id = execution_info.id if isinstance(execution_info, WorkflowExecutionInfo) else execution_info
        conditions.append(f"tags.{MlflowTag.MLFLOW_RUN_NAME.value} = '{execution_id}'")
    if additional_filters is not None:
        conditions.append(additional_filters)
    return " AND ".join(conditions)


def is_mlflow_using_local_artifact_storage(
    mlflow_run: Union[str, MlflowRun],
) -> bool:
    mlflow_run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MlflowRun) else mlflow_run
    mlflow_artifact_uri = urlparse(artifact_utils.get_artifact_uri(run_id=mlflow_run_id))  # type: ignore[no-untyped-call]
    return bool(mlflow_artifact_uri.scheme == "file")


def get_mlflow_local_artifact_storage_path(
    mlflow_run: Union[str, MlflowRun],
) -> Optional[Path]:
    mlflow_run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MlflowRun) else mlflow_run
    mlflow_artifact_uri = urlparse(artifact_utils.get_artifact_uri(run_id=mlflow_run_id))  # type: ignore[no-untyped-call]
    if mlflow_artifact_uri.scheme == "file":
        return Path(mlflow_artifact_uri.path)
    return None


def fetch_child_mlflow_runs(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    mlflow_run: Union[str, MlflowRun],
) -> Iterator[MlflowRun]:
    experiment = get_mlflow_experiment(experiment)
    mlflow_run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MlflowRun) else mlflow_run
    page_token: Optional[str] = None
    while True:
        runs = client.search_runs(
            experiment_ids=[experiment.experiment_id],
            filter_string=f"tags.{MlflowTag.MLFLOW_PARENT_RUN_ID.value} = '{mlflow_run_id}'",
            page_token=page_token,
        )
        yield from iter(runs)
        if runs.token is None:
            break
        page_token = runs.token


def update_mlflow_tags(
    client: MlflowClient,
    run: Union[str, MlflowRun],
    tags: dict[MlflowTag, str],
) -> None:
    run_id = run.info.run_id if isinstance(run, MlflowRun) else run
    for tag, value in tags.items():
        client.set_tag(run_id=run_id, key=tag.value, value=value)


def fetch_mlflow_runs(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    *,
    run_type: Optional[WorkflowRunType] = None,
    step_info: Optional[Union[str, WorkflowStepInfo]] = None,
    execution_info: Optional[Union[str, WorkflowExecutionInfo]] = None,
    with_children: bool = False,
) -> Iterator[MlflowRun]:
    experiment = get_mlflow_experiment(experiment)

    page_token: Optional[str] = None
    filter_string = build_filter_string(run_type, step_info, execution_info)
    while True:
        runs = client.search_runs(
            experiment_ids=[experiment.experiment_id],
            filter_string=filter_string,
            page_token=page_token,
        )
        for run in runs:
            yield run
            if with_children:
                yield from fetch_child_mlflow_runs(client, experiment, run)
        if runs.token is None:
            break
        page_token = runs.token


def fetch_mlflow_run(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    *,
    run_type: Optional[WorkflowRunType] = None,
    step_info: Optional[Union[str, WorkflowStepInfo]] = None,
    execution_info: Optional[Union[str, WorkflowExecutionInfo]] = None,
) -> Optional[MlflowRun]:
    return next(
        fetch_mlflow_runs(
            client,
            experiment,
            run_type=run_type,
            step_info=step_info,
            execution_info=execution_info,
        ),
        None,
    )


def generate_new_execution_id(client: MlflowClient, experiment: Union[str, MlflowExperiment]) -> WorkflowExecutionID:
    experiment = get_mlflow_experiment(experiment)
    while True:
        execution_id = uuid.uuid4().hex[:8]
        if fetch_mlflow_run(client, experiment, execution_info=execution_id) is None:
            return WorkflowExecutionID(execution_id)


def add_mlflow_run(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    step_or_execution_info: Union[WorkflowStepInfo, WorkflowExecutionInfo],
    parent_run_id: Optional[str] = None,
) -> MlflowRun:
    experiment = get_mlflow_experiment(experiment)

    run_name: str
    params: MlflowParams
    tags: dict[MlflowTag, str]

    if isinstance(step_or_execution_info, WorkflowStepInfo):
        run_name = step_or_execution_info.name
        params = get_step_params(step_or_execution_info)
        tags = get_step_tags(step_or_execution_info)
    elif isinstance(step_or_execution_info, WorkflowExecutionInfo):
        assert step_or_execution_info.id is not None
        run_name = step_or_execution_info.id
        params = get_execution_params(step_or_execution_info)
        tags = get_execution_tags(step_or_execution_info)
    else:
        raise ValueError(f"Unsupported type: {type(step_or_execution_info)}")

    if parent_run_id is not None:
        tags[MlflowTag.MLFLOW_PARENT_RUN_ID] = parent_run_id

    run = client.create_run(
        experiment_id=experiment.experiment_id,
        run_name=run_name,
        tags=context_registry.resolve_tags({tag.value: value for tag, value in tags.items()}),
    )
    for key, value in params.items():
        client.log_param(run.info.run_id, key, value)

    return run


def download_mlflow_artifacts(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    step_or_execution_info: Union[WorkflowStepInfo, WorkflowExecutionInfo],
    directory: Union[str, PathLike],
    artifact_path: Optional[str] = None,
) -> None:
    run_type = (
        WorkflowRunType.STEP if isinstance(step_or_execution_info, WorkflowStepInfo) else WorkflowRunType.EXECUTION
    )
    step_info = step_or_execution_info if isinstance(step_or_execution_info, WorkflowStepInfo) else None
    execution_info = step_or_execution_info if isinstance(step_or_execution_info, WorkflowExecutionInfo) else None
    run = fetch_mlflow_run(client, experiment, run_type=run_type, step_info=step_info, execution_info=execution_info)
    if run is None:
        raise FileNotFoundError("Run not found")
    with tempfile.TemporaryDirectory() as temp_dir:
        mlflow.artifacts.download_artifacts(
            run_id=run.info.run_id,
            artifact_path=artifact_path,
            dst_path=temp_dir,
        )
        Path(temp_dir).replace(Path(directory))


def upload_mlflow_artifacts(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    step_or_execution_info: Union[WorkflowStepInfo, WorkflowExecutionInfo],
    directory: Union[str, PathLike],
    artifact_path: Optional[str] = None,
) -> None:
    run = fetch_mlflow_run(
        client,
        experiment,
        step_info=step_or_execution_info if isinstance(step_or_execution_info, WorkflowStepInfo) else None,
        execution_info=step_or_execution_info if isinstance(step_or_execution_info, WorkflowExecutionInfo) else None,
    )
    if run is None:
        raise ValueError("Run not found")
    client.log_artifacts(
        run_id=run.info.run_id,
        local_dir=str(directory),
        artifact_path=artifact_path,
    )


def terminate_mlflow_run(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    step_or_execution_state: Union[WorkflowStepState, WorkflowExecutionState],
) -> None:
    step_info: Optional[str] = None
    execution_info: Optional[str] = None
    run_status: MlflowRunStatus
    end_time: Optional[int] = None
    if isinstance(step_or_execution_state, WorkflowStepState):
        if step_or_execution_state.status not in (
            WorkflowStepStatus.COMPLETED,
            WorkflowStepStatus.FAILURE,
            WorkflowStepStatus.CANCELED,
        ):
            raise ValueError(f"Invalid step status: {step_or_execution_state.status}")
        step_info = step_or_execution_state.fingerprint
        run_status = MlflowRunStatus.from_step_sttus(step_or_execution_state.status)
        end_time = (
            int(step_or_execution_state.finished_at.timestamp() * 1000) if step_or_execution_state.finished_at else None
        )
    elif isinstance(step_or_execution_state, WorkflowExecutionState):
        if step_or_execution_state.status not in (
            WorkflowExecutionStatus.COMPLETED,
            WorkflowExecutionStatus.FAILURE,
            WorkflowExecutionStatus.CANCELED,
        ):
            raise ValueError(f"Invalid execution status: {step_or_execution_state.status}")
        execution_info = step_or_execution_state.execution_id
        run_status = MlflowRunStatus.from_execution_status(step_or_execution_state.status)
        end_time = (
            int(step_or_execution_state.finished_at.timestamp() * 1000) if step_or_execution_state.finished_at else None
        )
    experiment = get_mlflow_experiment(experiment)
    run = fetch_mlflow_run(
        client,
        experiment,
        step_info=step_info,
        execution_info=execution_info,
    )
    if run is None:
        raise ValueError("Run not found")
    client.set_terminated(
        run_id=run.info.run_id,
        status=run_status.value,
        end_time=end_time,
    )


def remove_mlflow_run(
    client: MlflowClient,
    experiment: Union[str, MlflowExperiment],
    step_or_execution_info: Union[WorkflowStepInfo, WorkflowExecutionInfo],
) -> None:
    run = fetch_mlflow_run(
        client,
        experiment,
        step_info=step_or_execution_info if isinstance(step_or_execution_info, WorkflowStepInfo) else None,
        execution_info=step_or_execution_info if isinstance(step_or_execution_info, WorkflowExecutionInfo) else None,
    )
    if run is None:
        raise ValueError("Run not found")
    client.delete_run(run.info.run_id)


def get_mlflow_experiment(experiment: Union[str, MlflowExperiment]) -> MlflowExperiment:
    if isinstance(experiment, str):
        client = MlflowClient()
        experiment_or_none = client.get_experiment_by_name(experiment)
        if experiment_or_none is None:
            mlflow.create_experiment(experiment)
            experiment_or_none = client.get_experiment_by_name(experiment)
            assert experiment_or_none is not None
        experiment = experiment_or_none
    return experiment


def get_step_params(step_info: WorkflowStepInfo) -> MlflowParams:
    config = as_jsonvalue(step_info.step.config)
    assert isinstance(config, dict)
    return flatten_params(config)


def get_execution_params(execution_info: WorkflowExecutionInfo) -> MlflowParams:
    config = as_jsonvalue(execution_info.graph.to_dict())
    assert isinstance(config, dict)
    return flatten_params(config)


def get_step_tags(step_info: WorkflowStepInfo) -> dict[MlflowTag, str]:
    return {
        MlflowTag.MLFACTORY_RUN_TYPE: WorkflowRunType.STEP.value,
        MlflowTag.MLFACTORY_STEP_FINGERPRINT: step_info.fingerprint,
    }


def get_execution_tags(execution_info: WorkflowExecutionInfo) -> dict[MlflowTag, str]:
    assert execution_info.id is not None
    return {
        MlflowTag.MLFLOW_RUN_NAME: execution_info.id,
        MlflowTag.MLFACTORY_RUN_TYPE: WorkflowRunType.EXECUTION.value,
    }



from typing import Final, Union

_ALPHABET: Final[bytes] = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def b58encode(s: Union[str, bytes]) -> bytes:
    if isinstance(s, str):
        s = s.encode("ascii")
    original_length = len(s)
    s = s.lstrip(b"\0")
    stripped_length = len(s)
    n = int.from_bytes(s, "big")
    res = bytearray()
    while n:
        n, r = divmod(n, 58)
        res.append(_ALPHABET[r])
    return _ALPHABET[0:1] * (original_length - stripped_length) + bytes(reversed(res))


def b58decode(v: Union[str, bytes]) -> bytes:
    v = v.rstrip()
    if isinstance(v, str):
        v = v.encode("ascii")
    original_length = len(v)
    v = v.lstrip(_ALPHABET[0:1])
    stripped_length = len(v)
    n = 0
    for c in v:
        n = n * 58 + _ALPHABET.index(c)
    return n.to_bytes(original_length - stripped_length + (n.bit_length() + 7) // 8, "big")



from .cache import FilesystemWorkflowCache, MemoryWorkflowCache, WorkflowCache  # noqa: F401
from .callback import MultiWorkflowCallback, WorkflowCallback  # noqa: F401
from .constants import WORKFLOW_DEFAULT_DIRECTORY, WORKFLOW_DEFAULT_SETTINGS_PATH  # noqa: F401
from .executor import (  # noqa: F401
    DefaultWorkflowExecutor,
    WorkflowExecutionContext,
    WorkflowExecutionID,
    WorkflowExecutionInfo,
    WorkflowExecutionMetadata,
    WorkflowExecutionState,
    WorkflowExecutionStatus,
    WorkflowExecutor,
    use_execution_context,
)
from .format import Format, JsonFormat, PickleFormat  # noqa: F401
from .graph import WorkflowGraph  # noqa: F401
from .organizer import FilesystemWorkflowOrganizer, MemoryWorkflowOrganizer, WorkflowOrganizer  # noqa: F401
from .settings import WorkflowSettings  # noqa: F401
from .step import (  # noqa: F401
    WorkflowStep,
    WorkflowStepContext,
    WorkflowStepInfo,
    WorkflowStepState,
    WorkflowStepStatus,
    get_step_logger_from_info,
    step,
    use_step_context,
    use_step_logger,
)



import copy
import itertools
import json
import os
from collections.abc import Iterable, Mapping
from os import PathLike
from typing import Any, ClassVar, Optional, TypeVar, Union, cast

import colt
from rjsonnet import evaluate_file, evaluate_snippet

T = TypeVar("T", dict, list)


def _is_encodable(value: str) -> bool:
    return (value == "") or (value.encode("utf-8", "ignore") != b"")


def _environment_variables() -> dict[str, str]:
    return {key: value for key, value in os.environ.items() if _is_encodable(value)}


def _parse_overrides(serialized_overrides: str, ext_vars: Optional[dict[str, Any]] = None) -> dict[str, Any]:
    if serialized_overrides:
        ext_vars = {**_environment_variables(), **(ext_vars or {})}
        output = json.loads(evaluate_snippet("", serialized_overrides, ext_vars=ext_vars))
        assert isinstance(output, dict), "Overrides must be a JSON object."
        return output
    else:
        return {}


def _with_overrides(original: T, overrides_dict: dict[str, Any], prefix: str = "") -> T:
    merged: T
    keys: Union[Iterable[str], Iterable[int]]
    if isinstance(original, list):
        merged = [None] * len(original)
        keys = cast(Iterable[int], range(len(original)))
    elif isinstance(original, dict):
        merged = {}
        keys = cast(
            Iterable[str],
            itertools.chain(original.keys(), (k for k in overrides_dict if "." not in k and k not in original)),
        )
    else:
        if prefix:
            raise ValueError(
                f"overrides for '{prefix[:-1]}.*' expected list or dict in original, " f"found {type(original)} instead"
            )
        else:
            raise ValueError(f"expected list or dict, found {type(original)} instead")

    used_override_keys: set[str] = set()
    for key in keys:
        if str(key) in overrides_dict:
            merged[key] = copy.deepcopy(overrides_dict[str(key)])
            used_override_keys.add(str(key))
        else:
            overrides_subdict = {}
            for o_key in overrides_dict:
                if o_key.startswith(f"{key}."):
                    overrides_subdict[o_key[len(f"{key}.") :]] = overrides_dict[o_key]
                    used_override_keys.add(o_key)
            if overrides_subdict:
                merged[key] = _with_overrides(original[key], overrides_subdict, prefix=prefix + f"{key}.")
            else:
                merged[key] = copy.deepcopy(original[key])

    unused_override_keys = [prefix + key for key in set(overrides_dict.keys()) - used_override_keys]
    if unused_override_keys:
        raise ValueError(f"overrides dict contains unused keys: {unused_override_keys}")

    return merged


def load_jsonnet(
    filename: Union[str, PathLike],
    ext_vars: Optional[Mapping[str, Any]] = None,
    overrides: Optional[str] = None,
) -> Any:
    ext_vars = {**_environment_variables(), **(ext_vars or {})}
    output = json.loads(evaluate_file(str(filename), ext_vars=ext_vars))
    if overrides:
        output = _with_overrides(output, _parse_overrides(overrides, ext_vars=ext_vars))
    return output


_T_FromJsonnet = TypeVar("_T_FromJsonnet", bound="FromJsonnet")


class FromJsonnet:
    __COLT_BUILDER__: ClassVar = colt.ColtBuilder(typekey="type")

    @classmethod
    def from_jsonnet(
        cls: type[_T_FromJsonnet],
        filename: Union[str, PathLike],
        ext_vars: Optional[Mapping[str, Any]] = None,
        overrides: Optional[str] = None,
    ) -> _T_FromJsonnet:
        config = load_jsonnet(filename, ext_vars=ext_vars, overrides=overrides)
        config = cls.__pre_init__(config)
        obj: _T_FromJsonnet = cls.__COLT_BUILDER__(config, cls)
        setattr(obj, "__config__", config)
        return obj

    @classmethod
    def __pre_init__(self, config: Any) -> Any:
        return config

    def to_dict(self) -> Any:
        return getattr(self, "__config__")



from typing import Final

from harbory.workflow.constants import WORKFLOW_INTEGRATION_DIRECTORY

DEFAULT_MLFLOW_EXPERIMENT_NAME: Final = "Default"
DEFAULT_MLFLOW_DIRECTORY: Final = WORKFLOW_INTEGRATION_DIRECTORY / "mlflow"



import dataclasses

from .executor import DefaultWorkflowExecutor, WorkflowExecutor
from .organizer import FilesystemWorkflowOrganizer, WorkflowOrganizer


def _default_executor() -> WorkflowExecutor:
    return DefaultWorkflowExecutor()


def _default_organizer() -> WorkflowOrganizer:
    return FilesystemWorkflowOrganizer()


@dataclasses.dataclass(frozen=True)
class WorkflowSettings:
    executor: WorkflowExecutor = dataclasses.field(default_factory=_default_executor)
    organizer: WorkflowOrganizer = dataclasses.field(default_factory=_default_organizer)



from collections.abc import Sequence
from typing import TYPE_CHECKING

from colt import Registrable

if TYPE_CHECKING:
    from .executor import WorkflowExecutionContext
    from .step import WorkflowStepContext


class WorkflowCallback(Registrable):
    def on_execution_start(
        self,
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        pass

    def on_execution_end(
        self,
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        pass

    def on_step_start(
        self,
        step_context: "WorkflowStepContext",
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        pass

    def on_step_end(
        self,
        step_context: "WorkflowStepContext",
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        pass


@WorkflowCallback.register("empty")
class EmptyWorkflowCallback(WorkflowCallback): ...


@WorkflowCallback.register("multi")
class MultiWorkflowCallback(WorkflowCallback):
    def __init__(self, callbacks: Sequence["WorkflowCallback"]) -> None:
        self._callbacks = callbacks

    def on_execution_start(
        self,
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        for callback in self._callbacks:
            callback.on_execution_start(execution_context)

    def on_execution_end(
        self,
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        for callback in self._callbacks:
            callback.on_execution_end(execution_context)

    def on_step_start(
        self,
        step_context: "WorkflowStepContext",
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        for callback in self._callbacks:
            callback.on_step_start(step_context, execution_context)

    def on_step_end(
        self,
        step_context: "WorkflowStepContext",
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        for callback in self._callbacks:
            callback.on_step_end(step_context, execution_context)



import sys
from collections.abc import Callable, Hashable
from typing import Generic, TextIO, TypeVar

T_Node = TypeVar("T_Node", bound=Hashable)


class DAG(Generic[T_Node]):
    def __init__(self, dependencies: dict[T_Node, set[T_Node]]) -> None:
        nodes = set(dependencies.keys() | set(node for deps in dependencies.values() for node in deps))
        empty_dependencies: dict[T_Node, set[T_Node]] = {node: set() for node in nodes}

        self._dependencies = {**empty_dependencies, **dependencies}

    def __len__(self) -> int:
        return len(self._dependencies)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DAG):
            return NotImplemented
        return self._dependencies == other._dependencies

    def __hash__(self) -> int:
        return hash(frozenset((node, frozenset(deps)) for node, deps in self._dependencies.items()))

    @property
    def nodes(self) -> set[T_Node]:
        return set(self._dependencies.keys())

    def subgraph(self, nodes: set[T_Node]) -> "DAG":
        return DAG({node: nodes & deps for node, deps in self._dependencies.items() if node in nodes})

    def in_degree(self, node: T_Node) -> int:
        return len(self._dependencies[node])

    def successors(self, node: T_Node) -> set[T_Node]:
        return {successor for successor, deps in self._dependencies.items() if node in deps}

    def weekly_connected_components(self) -> set["DAG"]:
        groups: list[set[T_Node]] = []

        for node, deps in self._dependencies.items():
            current_group = {node} | deps
            for group in groups:
                if group & current_group:
                    group.update(current_group)
                    break
            else:
                groups.append(current_group)

        for i, group in enumerate(groups):
            for other_group in groups[i + 1 :]:
                if group & other_group:
                    group.update(other_group)
                    other_group.clear()

        groups = [group for group in groups if group]

        return {DAG({node: self._dependencies[node] for node in group}) for group in groups}

    def visualize(
        self,
        *,
        indent: int = 2,
        output: TextIO = sys.stdout,
        rename: Callable[[T_Node], str] = str,
    ) -> None:
        locations: dict[T_Node, tuple[int, int]] = {}

        def _process_dag(dag: DAG, level: int) -> None:
            for subdag in sorted(dag.weekly_connected_components(), key=lambda g: sorted(g.nodes)):
                _process_component(subdag, level)

        def _process_component(dag: DAG, level: int) -> None:
            sources = set(node for node in dag.nodes if dag.in_degree(node) == 0)
            for i, node in enumerate(sorted(sources)):
                if node in locations:
                    raise ValueError(f"Node {node} already placed")
                locations[node] = (len(locations), level + i)

            subdag = dag.subgraph(dag.nodes - sources)
            _process_dag(subdag, level + len(sources))

        _process_dag(self, 0)

        a = [[" "] * level * indent for _, level in sorted(locations.values())]
        for node, (position, level) in sorted(locations.items(), key=lambda x: x[1][0]):
            successors = sorted(self.successors(node), key=lambda v: locations[v][0])
            if not successors:
                continue

            last_successor_position, last_successor_level = locations[successors[-1]]
            for successor_position in range(position + 1, last_successor_position):
                a[successor_position][level * indent] = "│"
            for successor in successors[:-1]:
                successor_position, successor_level = locations[successor]
                a[successor_position][level * indent] = (
                    "┼" if level > 0 and a[successor_position][level * indent - 1] == "─" else "├"
                )
                for offset in range(level * indent + 1, successor_level * indent):
                    a[successor_position][offset] = "─"
            a[last_successor_position][level * indent] = (
                "┴" if level > 0 and a[last_successor_position][level * indent - 1] == "─" else "╰"
            )
            for offset in range(level * indent + 1, last_successor_level * indent):
                a[last_successor_position][offset] = "─"

        output.writelines(
            "".join(a[position]) + "• " + rename(node) + "\n"
            for node, (position, _) in sorted(locations.items(), key=lambda x: x[1][0])
        )



import json
import logging
from types import TracebackType
from typing import Generic, Literal, Optional, TextIO, Type, TypeVar, Union

T_TextIO = TypeVar("T_TextIO", bound=TextIO)


class JsonFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        log_regord = {
            "time": self.formatTime(record),
            "name": record.name,
            "level": record.levelname,
            "message": record.getMessage(),
        }
        return json.dumps(log_regord, ensure_ascii=False)


class LogCapture(Generic[T_TextIO]):
    def __init__(
        self,
        file: T_TextIO,
        logger: Optional[Union[str, logging.Logger]] = None,
        formatter: Optional[logging.Formatter] = None,
    ) -> None:
        formatter = formatter or JsonFormatter()
        logger = logging.getLogger(logger) if isinstance(logger, str) else (logger or logging.getLogger())

        self._file = file
        self._logger = logger
        self._formatter = formatter
        self._stream_handler: Optional[logging.StreamHandler] = None

    @property
    def stream(self) -> T_TextIO:
        return self._file

    def start(self) -> None:
        if not self._stream_handler:
            self._stream_handler = logging.StreamHandler(self._file)
            self._stream_handler.setFormatter(self._formatter)
            self._logger.addHandler(self._stream_handler)

    def stop(self) -> None:
        if self._stream_handler:
            self._logger.removeHandler(self._stream_handler)
            self._stream_handler = None

    def __enter__(self) -> None:
        self.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        self.stop()
        return False



import argparse
import pathlib
import sys
from logging import getLogger

from harbory.settings import load_harbory_settings
from harbory.workflow import WorkflowExecutionID, WorkflowExecutionInfo, WorkflowGraph

from .subcommand import Subcommand

logger = getLogger(__name__)


@Subcommand.register("workflow")
class WorkflowCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "--settings",
            type=str,
            default=None,
            help="workflow environment settings file path",
        )


@WorkflowCommand.register("run")
class WorkflowRunCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "config",
            type=str,
            help="config file path",
        )
        self.parser.add_argument(
            "--execution-id",
            type=WorkflowExecutionID,
            default=None,
            help="execution id",
        )
        self.parser.add_argument(
            "--step",
            type=str,
            default=None,
            help="step name to execute",
        )
        self.parser.add_argument(
            "--overrides",
            type=str,
            default=None,
            help="overrides jsonnet file path",
        )

    def run(self, args: argparse.Namespace) -> None:
        harbory_settings = load_harbory_settings(args.settings)
        settings = harbory_settings.workflow
        config_path = pathlib.Path(args.config)

        logger.info(f"Load workflow from {config_path}")
        graph = WorkflowGraph.from_jsonnet(config_path, overrides=args.overrides)
        if args.step is not None:
            graph = graph.get_subgraph(args.step)

        execution = WorkflowExecutionInfo(graph, id=args.execution_id)

        organizer = settings.organizer
        executor = settings.executor

        if execution.id is not None and organizer.exists(execution.id):
            logger.error(f"Execution {execution.id} already exists")
            sys.exit(1)

        organizer.run(executor, execution)


@WorkflowCommand.register("remove")
class WorkflowRemoveCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "execution_id",
            type=WorkflowExecutionID,
            help="execution id",
        )

    def run(self, args: argparse.Namespace) -> None:
        harbory_settings = load_harbory_settings(args.settings)
        settings = harbory_settings.workflow
        organizer = settings.organizer

        if not organizer.exists(args.execution_id):
            logger.error(f"Execution {args.execution_id} does not exist")
            sys.exit(1)

        organizer.remove(args.execution_id)



import importlib
import json
from collections.abc import Iterator
from pathlib import Path
from typing import IO, Any, ClassVar, Generic, Optional, TypeVar, Union, cast

import colt
import dill
from colt import Registrable
from pydantic import BaseModel

from harbory.types import DataContainer, IDataclass, INamedTuple, JsonValue

from .utils import WorkflowJSONEncoder

_JsonFormattable = Union[DataContainer, JsonValue]

_S = TypeVar("_S")
_T = TypeVar("_T")
_T_JsonFormattable = TypeVar("_T_JsonFormattable", bound=Union[_JsonFormattable, Iterator[_JsonFormattable]])


class Format(Generic[_T], Registrable):
    @property
    def identifier(self) -> str:
        return f"{self.__class__.__module__}:{self.__class__.__name__}"

    def write(self, artifact: _T, directory: Path) -> None:
        raise NotImplementedError

    def read(self, directory: Path) -> _T:
        raise NotImplementedError

    @classmethod
    def is_default_of(cls, obj: Any) -> bool:
        return False


@Format.register("pickle")
class PickleFormat(Format[_T], Generic[_T]):
    class _IteratorWrapper(Generic[_S]):
        def __init__(self, path: Path) -> None:
            self._file: Optional[IO[Any]] = path.open("rb")
            self._unpickler = dill.Unpickler(self._file)

        def __iter__(self) -> Iterator[_S]:
            return self

        def __next__(self) -> _S:
            if self._file is None:
                raise StopIteration
            try:
                return cast(_S, self._unpickler.load())
            except EOFError:
                self._file.close()
                self._file = None
                raise StopIteration

    def _get_artifact_path(self, directory: Path) -> Path:
        return directory / "artifact.pkl"

    def write(self, artifact: _T, directory: Path) -> None:
        artifact_path = self._get_artifact_path(directory)
        with open(artifact_path, "wb") as f:
            pickler = dill.Pickler(f)
            if isinstance(artifact, Iterator):
                pickler.dump(True)
                for item in artifact:
                    pickler.dump(item)
            else:
                pickler.dump(False)
                pickler.dump(artifact)

    def read(self, directory: Path) -> _T:
        artifact_path = self._get_artifact_path(directory)
        with open(artifact_path, "rb") as f:
            unpickler = dill.Unpickler(f)
            is_iterator = unpickler.load()
            if is_iterator:
                return cast(_T, self._IteratorWrapper(artifact_path))
            return cast(_T, unpickler.load())


@Format.register("json")
class JsonFormat(Format[_T_JsonFormattable], Generic[_T_JsonFormattable]):
    class _IteratorWrapper(Generic[_S]):
        def __init__(self, path: Path, artifact_class: Optional[type[_S]]) -> None:
            self._file = path.open("r")
            self._artifact_class = artifact_class

        def __iter__(self) -> Iterator[_S]:
            return self

        def __next__(self) -> _S:
            line = self._file.readline()
            if not line:
                self._file.close()
                raise StopIteration
            data = json.loads(line)
            if self._artifact_class is not None:
                return colt.build(data, self._artifact_class)
            return cast(_S, data)

    def write(self, artifact: _T_JsonFormattable, directory: Path) -> None:
        artifact_class: Optional[type[_T_JsonFormattable]] = None
        if isinstance(artifact, Iterator):
            artifact_path = directory / "artifact.jsonl"
            with open(artifact_path, "w") as f:
                for item in artifact:
                    artifact_class = artifact_class or type(item)
                    json.dump(item, f, cls=WorkflowJSONEncoder, ensure_ascii=False)
                    f.write("\n")
        else:
            artifact_class = type(artifact)
            artifact_path = directory / "artifact.json"
            with open(artifact_path, "w") as f:
                json.dump(artifact, f, cls=WorkflowJSONEncoder, ensure_ascii=False)
        if artifact_class is not None:
            metadata = {"module": artifact_class.__module__, "class": artifact_class.__name__}
            metadata_path = directory / "metadata.json"
            with open(metadata_path, "w") as f:
                json.dump(metadata, f, ensure_ascii=False)

    def read(self, directory: Path) -> _T_JsonFormattable:
        metadata_path = directory / "metadata.json"
        artifact_class: Optional[type[_T_JsonFormattable]] = None
        if metadata_path.exists():
            with open(metadata_path, "r") as f:
                metadata = json.load(f)
            module = importlib.import_module(metadata["module"])
            artifact_class = getattr(module, metadata["class"])

        is_iterator = (directory / "artifact.jsonl").exists()
        if is_iterator:
            artifact_path = directory / "artifact.jsonl"
            return cast(_T_JsonFormattable, self._IteratorWrapper(artifact_path, artifact_class))

        artifact_path = directory / "artifact.json"
        with open(artifact_path, "r") as f:
            data = json.load(f)
            if artifact_class is not None:
                return colt.build(data, artifact_class)
            return cast(_T_JsonFormattable, data)

    @classmethod
    def is_default_of(cls, obj: Any) -> bool:
        return isinstance(obj, (int, float, str, bool, dict, list, tuple, IDataclass, INamedTuple, BaseModel))


@Format.register("auto")
class AutoFormat(Format[_T]):
    _DEFAULT_FORMAT: ClassVar[str] = "pickle"
    _FORMAT_FILENAME: ClassVar[str] = "__format__"

    @classmethod
    def _get_default_format_name(cls, obj: _T) -> str:
        registry = Format._registry[Format]
        # NOTE: `reversed` is a workaround to prioritize the last registered format
        # that may be more specific than the first registered format
        for name, (format_cls, _) in reversed(registry.items()):
            if format_cls.is_default_of(obj):
                return name
        return cls._DEFAULT_FORMAT

    def write(self, artifact: _T, directory: Path) -> None:
        format_name = self._get_default_format_name(artifact)
        format = cast(type[Format[_T]], Format.by_name(format_name))()
        format.write(artifact, directory)
        (directory / self._FORMAT_FILENAME).write_text(json.dumps({"name": format_name}))

    def read(self, directory: Path) -> _T:
        format_metadata = json.loads((directory / self._FORMAT_FILENAME).read_text())
        format_name = format_metadata["name"]
        format = cast(type[Format[_T]], Format.by_name(format_name))()
        return format.read(directory)



from pathlib import Path
from typing import Final

COLT_TYPEKEY: Final = "type"
COLT_ARGSKEY: Final = "*"

DEFAULT_WORKING_DIRECTORY: Final = Path.cwd()
DEFAULT_MLFACTORY_DIRECTORY: Final = DEFAULT_WORKING_DIRECTORY / ".harbory"
DEFAULT_MLFACTORY_SETTINGS_PATH: Final = DEFAULT_WORKING_DIRECTORY / "harbory.yml"



import dataclasses
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence, Sized
from typing import Any, Generic, Optional, TypeVar, cast

import colt
from mpire import WorkerPool

from harbory.common.iterutils import SizedIterator, batched
from harbory.types import BaseModel, IDataclass, INamedTuple, T_DataContainer

# inputs/outputs
S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
# fixtures
E = TypeVar("E")
F = TypeVar("F")
# runtime params
P = TypeVar("P")
Q = TypeVar("Q")


class Processor(Generic[S, T, F, P], colt.Registrable):
    """
    A base class for processors.

    To define a processor, inherit this class and implement the `apply` method, which
    applies the processor to a single input. The `apply_batch` method is implemented
    in terms of `apply`, and can be overridden for efficiency.

    Here is an example of processor that tokenizes a string by splitting on spaces:

    Example:
        >>> class MyProcessor(Processor[str, list[str], None, None]):
        >>>     fixtures = None
        >>>
        >>>     def apply_batch(self, inputs: list[str], fixtures: None, params: None) -> list[str]:
        >>>         return [x.split() for x in inputs]

    To apply the processor to a sequence of inputs, call the processor object as a
    function. The `batch_size` argument controls the number of inputs to process
    at a time. The `max_workers` argument controls the number of threads to use
    for multi-thread processing. For example, to apply the processor to a sequence
    of inputs, 100 at a time, using 4 threads, do the following:

    Example:
        >>> processor = MyProcessor()
        >>> inputs = ["This is a test.", "This is another test.", ...]
        >>> outputs = processor(inputs, batch_size=100, max_workers=4)

    Chaining processors together is done using the `|` operator. For example, to
    create a processor consisting of three steps, `first_step`, `second_step`, and
    `third_step`, do the following:

    Example:
        >>> first_step = FirstProcessor()
        >>> second_step = SecondProcessor()
        >>> third_step = ThirdProcessor()
        >>> processor = first_step | second_step | third_step

    Args:
        batch_size: The batch size. Defaults to `1`.
        max_workers: The maximum number of workers to use for multi-thread
            processing. Defaults to `1`.
    """

    def __init__(
        self,
        *,
        batch_size: int = 1,
        max_workers: int = 1,
    ) -> None:
        if batch_size < 1:
            raise ValueError("batch_size must be at least 1")
        if max_workers < 1:
            raise ValueError("max_workers must be at least 1")

        self._batch_size = batch_size
        self._max_workers = max_workers

    @property
    def fixtures(self) -> F:
        raise NotImplementedError

    def apply_batch(
        self,
        batch: Sequence[S],
        fixtures: F,
        params: Optional[P] = None,
    ) -> list[T]:
        """
        Apply the processor to a batch of inputs.

        Args:
            batch: The batch of inputs.
            fixtures: The fixtures to use.
            params: The runtime parameters.

        Returns:
            The batch of outputs.
        """

        raise NotImplementedError

    def apply(
        self,
        x: S,
        params: Optional[P] = None,
    ) -> T:
        """
        Apply the processor to a single input.

        Args:
            x: The input.
            params: The runtime parameters.

        Returns:
            The output.
        """

        return self.apply_batch([x], self.fixtures, params)[0]

    def __call__(
        self,
        inputs: Iterable[S],
        params: Optional[P] = None,
        *,
        batch_size: Optional[int] = None,
        max_workers: Optional[int] = None,
    ) -> Iterator[T]:
        """
        Apply the processor to a sequence of inputs.

        Args:
            inputs: The sequence of inputs.
            params: The runtime parameters.
            batch_size: The batch size. Defaults to `1`.
            max_workers: The maximum number of workers to use for multi-thread
                processing. Defaults to `1`.

        Returns:
            An iterator over the outputs.
        """

        batch_size = batch_size or self._batch_size or 1
        max_workers = max_workers or self._max_workers or 1

        assert batch_size is not None
        assert max_workers is not None

        if batch_size < 1:
            raise ValueError("batch_size must be at least 1")
        if max_workers < 1:
            raise ValueError("max_workers must be at least 1")

        def iterator() -> Iterator[T]:
            assert batch_size is not None
            assert max_workers is not None

            if max_workers < 2:
                fixtures = self.fixtures
                for batch in batched(inputs, batch_size):
                    yield from self.apply_batch(batch, fixtures, params)
            else:

                def apply_batch(shared_objects: tuple[F, Optional[P]], *batch: S) -> list[T]:
                    fixtures, params = shared_objects
                    return self.apply_batch(batch, fixtures, params)

                with WorkerPool(
                    n_jobs=max_workers,
                    shared_objects=(self.fixtures, params),
                ) as pool:
                    for results in pool.imap(
                        apply_batch,
                        batched(inputs, batch_size),
                    ):
                        yield from results

        if isinstance(inputs, Sized):
            return SizedIterator(iterator(), len(inputs))

        return iterator()

    def __or__(self, other: "Processor[T, U, E, Q]") -> "Processor[S, U, tuple[F, E], tuple[P, Q]]":
        return ComposeProcessor(self, other)

    @classmethod
    def from_callable(
        cls,
        func: Callable[[Sequence[S], F, Optional[Q]], list[T]],
        fixtures: F,
        **kwargs: Any,
    ) -> "Processor[S, T, F, Q]":
        """
        Create a processor from a callable.

        Args:
            func: The callable.
            fixtures: The fixtures to use.

        Returns:
            The processor.
        """

        return CallableProcessor(func, fixtures, **kwargs)


@Processor.register("compose")
class ComposeProcessor(Processor[S, U, tuple[E, F], tuple[P, Q]]):
    """
    A processor that is the composition of two processors.

    Args:
        first: The first processor.
        second: The second processor.
        batch_size: The batch size. Defaults to `1`.
        max_workers: The maximum number of workers to use for multi-thread
            processing. Defaults to `1`.
    """

    def __init__(
        self,
        first: Processor[S, T, E, P],
        second: Processor[T, U, F, Q],
        *,
        batch_size: int = 1,
        max_workers: int = 1,
    ):
        super().__init__(batch_size=batch_size, max_workers=max_workers)
        self.first = first
        self.second = second

    def apply_batch(
        self,
        batch: Sequence[S],
        fixtures: tuple[E, F],
        params: Optional[tuple[Optional[P], Optional[Q]]] = None,
    ) -> list[U]:
        params = params or (None, None)
        return self.second.apply_batch(self.first.apply_batch(batch, fixtures[0], params[0]), fixtures[1], params[1])

    def __call__(
        self,
        inputs: Iterable[S],
        params: Optional[tuple[Optional[P], Optional[Q]]] = None,
        *,
        batch_size: Optional[int] = None,
        max_workers: Optional[int] = None,
    ) -> Iterator[U]:
        params = params or (None, None)
        return self.second(
            self.first(
                inputs,
                params[0],
                batch_size=batch_size,
                max_workers=max_workers,
            ),
            params[1],
            batch_size=batch_size,
            max_workers=max_workers,
        )


@Processor.register("callable")
class CallableProcessor(Processor[S, T, F, P]):
    """
    A processor that can be created from a callable.

    Args:
        func: The callable.
        batch_size: The batch size. Defaults to `1`.
        max_workers: The maximum number of workers to use for multi-thread
            processing. Defaults to `1`.
    """

    def __init__(
        self,
        func: Callable[[Sequence[S], F, Optional[P]], list[T]],
        fixtures: F,
        *,
        batch_size: int = 1,
        max_workers: int = 1,
    ) -> None:
        super().__init__(batch_size=batch_size, max_workers=max_workers)
        self._func = func
        self._fixtures = fixtures

    @property
    def fixtures(self) -> F:
        return self._fixtures

    def apply_batch(
        self,
        batch: Sequence[S],
        fixtures: F,
        params: Optional[P] = None,
    ) -> list[T]:
        return self._func(batch, fixtures, params)


@Processor.register("chain")
class ChainProcessor(Processor[S, S, list[Any], list[Any]]):
    """
    A processor that is the composition of multiple processors.
    Note that each processor must have the same input and output types.

    Args:
        steps: The sequence of processors.
        batch_size: The batch size. Defaults to `1`.
        max_workers: The maximum number of workers to use for multi-thread
    """

    def __init__(
        self,
        steps: Sequence[Processor[S, S, Any, Any]],
        batch_size: int = 1,
        max_workers: int = 1,
    ) -> None:
        super().__init__(batch_size=batch_size, max_workers=max_workers)
        self._steps = steps or [PassThroughProcessor()]

    @property
    def fixtures(self) -> list[Any]:
        return [step.fixtures() for step in self._steps]

    def apply_batch(
        self,
        batch: Sequence[S],
        fixtures: list[Any],
        params: Optional[list[Any]] = None,
    ) -> list[S]:
        if params is not None and len(params) != len(self._steps):
            raise ValueError("params must have the same length as steps")
        params = params or [None] * len(self._steps)
        for s, f, p in zip(self._steps, fixtures, params):
            batch = s.apply_batch(batch, f, p)
        return list(batch)

    def __call__(
        self,
        inputs: Iterable[S],
        params: Optional[list[Any]] = None,
        *,
        batch_size: Optional[int] = None,
        max_workers: Optional[int] = None,
    ) -> Iterator[S]:
        if params is not None and len(params) != len(self._steps):
            raise ValueError("params must have the same length as steps")
        params = params or [None] * len(self._steps)
        output = self._steps[0](inputs, params[0], batch_size=batch_size, max_workers=max_workers)
        for step, param in zip(self._steps[1:], params[1:]):
            output = step(output, param, batch_size=batch_size, max_workers=max_workers)
        return output


@Processor.register("pass_through")
class PassThroughProcessor(Processor[S, S, None, None]):
    """
    A processor that does nothing.
    """

    fixtures = None

    def apply_batch(
        self,
        batch: Sequence[S],
        fixtures: None,
        params: None = None,
    ) -> list[S]:
        return list(batch)


@Processor.register("field")
class FieldProcessor(
    Processor[
        T_DataContainer,
        T_DataContainer,
        Mapping[str, Any],
        Mapping[str, Any],
    ]
):
    """
    A processor that applies a function to a field of a data container.
    """

    def __init__(
        self,
        processors: Mapping[str, Processor[Any, Any, Any, Any]],
        batch_size: int = 1,
        max_workers: int = 1,
    ) -> None:
        super().__init__(batch_size=batch_size, max_workers=max_workers)
        self._processors = processors

    @property
    def fixtures(self) -> Mapping[str, Any]:
        return {field: processor.fixtures for field, processor in self._processors.items()}

    def _get_field_value(self, item: T_DataContainer, field: str) -> Any:
        if isinstance(item, dict):
            return item[field]
        if isinstance(item, IDataclass):
            if field in item.__dataclass_fields__:
                return getattr(item, field)
            raise ValueError(f"Field '{field}' not found in data container")
        if isinstance(item, INamedTuple):
            if field in item._fields:
                return getattr(item, field)
            raise ValueError(f"Field '{field}' not found in data container")
        if isinstance(item, BaseModel):
            if field in item.model_fields:
                return getattr(item, field)
            raise ValueError(f"Field '{field}' not found in data container")
        raise ValueError("Data container must be a dict, a pydantic model or a dataclass")

    def _replace_field_value(self, item: T_DataContainer, field: str, value: Any) -> T_DataContainer:
        if isinstance(item, dict):
            return cast(T_DataContainer, {**item, field: value})
        if isinstance(item, IDataclass):
            return cast(T_DataContainer, dataclasses.replace(item, **{field: value}))
        if isinstance(item, INamedTuple):
            return cast(T_DataContainer, item._replace(**{field: value}))
        if isinstance(item, BaseModel):
            return cast(T_DataContainer, item.model_copy(update={field: value}))
        raise ValueError(f"Unsupported type: {type(item)}")

    def apply_batch(
        self,
        batch: Sequence[T_DataContainer],
        fixtures: Mapping[str, Any],
        params: Optional[Mapping[str, Any]] = None,
    ) -> list[T_DataContainer]:
        for key, processor in self._processors.items():
            field_fixtures = fixtures[key]
            field_params = params[key] if params is not None else None
            field_batch = [self._get_field_value(item, key) for item in batch]
            outputs = processor.apply_batch(field_batch, field_fixtures, field_params)
            batch = [self._replace_field_value(item, key, output) for item, output in zip(batch, outputs)]
        return list(batch)



import contextvars
import dataclasses
import datetime
import inspect
import typing
from collections.abc import Callable
from enum import Enum
from logging import Logger, getLogger
from typing import Any, ClassVar, Generic, Optional, TypeVar, Union, cast, overload

from colt import Lazy, Registrable

from harbory.common.astutils import normalize_source

from .format import AutoFormat, Format
from .types import StrictParamPath
from .utils import object_fingerprint

T = TypeVar("T")
T_Output = TypeVar("T_Output")
T_StepFunction = TypeVar("T_StepFunction", bound=Callable[..., Any])
T_WorkflowStep = TypeVar("T_WorkflowStep", bound="WorkflowStep")

_STEP_CONTEXT = contextvars.ContextVar[Optional["WorkflowStepContext"]]("_STEP_CONTEXT", default=None)


class WorkflowStepStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    FAILURE = "failure"
    CANCELED = "canceled"
    COMPLETED = "completed"


@dataclasses.dataclass(frozen=True)
class WorkflowStepState:
    fingerprint: str
    status: WorkflowStepStatus = WorkflowStepStatus.PENDING
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class WorkflowStepContext:
    info: "WorkflowStepInfo"
    state: WorkflowStepState


class WorkflowStep(Generic[T_Output], Registrable):
    VERSION: ClassVar[Optional[str]] = None
    DETERMINISTIC: ClassVar[bool] = True
    CACHEABLE: ClassVar[Optional[bool]] = None
    FORMAT: Format[T_Output]
    FUNCTION: Callable[..., T_Output]

    def __init__(self, *args: Any, **kwargs: Any):
        self._args = args
        self._kwargs = kwargs

    def __call__(self, context: Optional["WorkflowStepContext"]) -> T_Output:
        cls = cast(WorkflowStep[T_Output], self.__class__)
        ctx = contextvars.copy_context()

        def run() -> T_Output:
            if context is not None:
                _STEP_CONTEXT.set(context)
            return cls.FUNCTION(*self._args, **self._kwargs)

        return ctx.run(run)

    @classmethod
    def get_output_type(cls) -> type[T_Output]:
        return cast(type[T_Output], cls.FUNCTION.__annotations__["return"])

    @classmethod
    def from_callable(
        self,
        func: Callable[..., T_Output],
        *,
        version: Optional[str] = None,
        deterministic: bool = True,
        cacheable: Optional[bool] = None,
        format: Optional[Union[str, Format[T_Output]]] = None,
    ) -> type["WorkflowStep[T_Output]"]:
        if isinstance(format, str):
            format = cast(type[Format[T_Output]], Format.by_name(format))()

        class WrapperStep(WorkflowStep[T_Output]):
            VERSION = version
            DETERMINISTIC = deterministic
            CACHEABLE = cacheable
            FUNCTION = func
            FORMAT = cast(Format[T_Output], format or AutoFormat())

            def __init__(self, *args: Any, **kwargs: Any) -> None:
                super().__init__(*args, **kwargs)

        annotations = typing.get_type_hints(func)
        init_annotations = {k: v for k, v in annotations.items() if k != "return"}
        setattr(WrapperStep, "__name__", func.__name__)
        setattr(getattr(WrapperStep, "__init__"), "__annotations__", init_annotations)
        return WrapperStep

    @classmethod
    def get_source(cls) -> str:
        return inspect.getsource(cls.FUNCTION)

    @classmethod
    def get_normalized_source(cls) -> str:
        return normalize_source(cls.get_source())


@dataclasses.dataclass(frozen=True)
class WorkflowStepInfo(Generic[T_WorkflowStep]):
    name: str
    step: Lazy[T_WorkflowStep]
    dependencies: frozenset[tuple[StrictParamPath, "WorkflowStepInfo"]]

    @property
    def step_class(self) -> type[T_WorkflowStep]:
        step_class = self.step.constructor
        if not isinstance(step_class, type) or not issubclass(step_class, WorkflowStep):
            raise ValueError(f"Step {self.name} is not a subclass of WorkflowStep")
        return cast(type[T_WorkflowStep], step_class)

    @property
    def format(self) -> Format:
        return self.step_class.FORMAT

    @property
    def version(self) -> str:
        return self.step_class.VERSION or object_fingerprint(self.step_class.get_normalized_source())

    @property
    def deterministic(self) -> bool:
        return self.step_class.DETERMINISTIC

    @property
    def cacheable(self) -> Optional[bool]:
        return self.step_class.CACHEABLE

    @property
    def should_be_cached(self) -> bool:
        return self.cacheable or (self.cacheable is None and self.deterministic)

    @property
    def fingerprint(self) -> str:
        metadata = (self.name, self.version, self.deterministic, self.cacheable, self.format.identifier)
        config = self.step.config
        dependencies = frozenset(info.fingerprint for _, info in self.dependencies)
        return object_fingerprint((metadata, config, dependencies))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, WorkflowStepInfo):
            return False
        return self.fingerprint == other.fingerprint

    def __hash__(self) -> int:
        return hash(self.fingerprint)

    def __repr__(self) -> str:
        return f"WorkflowStepInfo[{self.name}:{self.fingerprint[:8]}]"

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "version": self.version,
            "format": self.format.identifier,
            "deterministic": self.deterministic,
            "cacheable": self.cacheable,
            "fingerprint": self.fingerprint,
            "config": self.step.config,
        }


@overload
def step(
    name: str,
    *,
    version: Optional[str] = ...,
    deterministic: bool = ...,
    cacheable: Optional[bool] = ...,
    exist_ok: bool = ...,
    format: Optional[Union[str, Format]] = ...,
) -> Callable[[T_StepFunction], T_StepFunction]: ...


@overload
def step(
    name: T_StepFunction,
    *,
    version: Optional[str] = ...,
    deterministic: bool = ...,
    cacheable: Optional[bool] = ...,
    exist_ok: bool = ...,
    format: Optional[Union[str, Format]] = ...,
) -> T_StepFunction: ...


def step(
    name: Union[str, T_StepFunction],
    *,
    version: Optional[str] = None,
    deterministic: bool = True,
    cacheable: Optional[bool] = None,
    exist_ok: bool = False,
    format: Optional[Union[str, Format]] = None,
) -> Union[T_StepFunction, Callable[[T_StepFunction], T_StepFunction]]:
    def register(name: str, func: T_StepFunction) -> None:
        step_class = WorkflowStep[Any].from_callable(
            func,
            version=version,
            deterministic=deterministic,
            cacheable=cacheable,
            format=format,
        )
        WorkflowStep.register(name, exist_ok=exist_ok)(step_class)

    def decorator(func: T_StepFunction) -> T_StepFunction:
        assert isinstance(name, str)
        register(name, func)
        return func

    if not isinstance(name, str):
        func = name
        register(func.__name__, func)
        return func

    return decorator


def use_step_context() -> Optional[WorkflowStepContext]:
    return _STEP_CONTEXT.get()


@overload
def use_step_logger(default: Union[str, Logger]) -> Logger: ...


@overload
def use_step_logger(default: None = ...) -> Optional[Logger]: ...


def use_step_logger(default: Optional[Union[str, Logger]] = None) -> Optional[Logger]:
    context = use_step_context()
    if context is not None:
        return get_step_logger_from_info(context.info)
    if default is None:
        return None
    if isinstance(default, str):
        return getLogger(default)
    return default


def get_step_logger_from_info(info: WorkflowStepInfo) -> Logger:
    return getLogger(f"harbory.workflow.step.{info.name}.{info.fingerprint[:8]}")



import dataclasses
import os
import subprocess
from os import PathLike
from pathlib import Path
from typing import Optional, Union


@dataclasses.dataclass(frozen=True)
class GitInfo:
    commit: str
    branch: str


class GitClient:
    @staticmethod
    def is_available() -> bool:
        return (
            subprocess.run(["git", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode == 0
        )

    def __init__(self, directory: Union[str, PathLike]) -> None:
        if not self.is_available():
            raise RuntimeError("Git is not available")

        self._directory = Path(directory)

    def root(self) -> Path:
        return Path(os.popen(f"git -C {self._directory} rev-parse --show-toplevel").read().strip())

    def is_initialized(self) -> bool:
        return (
            subprocess.run(
                ["git", "status"], cwd=self._directory, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            ).returncode
            == 0
        )

    def diff(self) -> str:
        return os.popen(f"git -C {self._directory} diff").read()

    def show(
        self,
        commit: Optional[str] = None,
        format: Optional[str] = None,
        no_patch: bool = False,
    ) -> str:
        command = f"git -C {self._directory} show"
        if commit:
            command += f" {commit}"
        if format:
            command += f" --format='{format}'"
        if no_patch:
            command += " --no-patch"
        return os.popen(command).read()

    def get_commit(self) -> str:
        return self.show(no_patch=True, format="%H").strip()

    def get_branch(self) -> str:
        return os.popen(f"git -C {self._directory} rev-parse --abbrev-ref HEAD").read().strip()

    def get_info(self) -> GitInfo:
        commit = self.get_commit()
        branch = self.get_branch()
        return GitInfo(commit, branch)


def get_git_info(directory: Optional[Union[str, PathLike]] = None) -> Optional[GitInfo]:
    if not GitClient.is_available():
        return None
    directory = directory or Path.cwd()
    client = GitClient(directory)
    if not client.is_available() or not client.is_initialized():
        return None
    return client.get_info()



from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, Union, cast

from colt import Registrable
from filelock import FileLock

if TYPE_CHECKING:
    from .step import WorkflowStep, WorkflowStepInfo


T = TypeVar("T")
T_WorkflowCache = TypeVar("T_WorkflowCache", bound="WorkflowCache")


class WorkflowCache(Registrable):
    def __getitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]") -> T:
        raise NotImplementedError

    def __setitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]", value: T) -> None:
        raise NotImplementedError

    def __delitem__(self, step_info: "WorkflowStepInfo") -> None:
        raise NotImplementedError

    def __contains__(self, step_info: "WorkflowStepInfo") -> bool:
        raise NotImplementedError


@WorkflowCache.register("empty")
class EmptyWorkflowCache(WorkflowCache):
    def __getitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]") -> T:
        raise KeyError(step_info)

    def __setitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]", value: T) -> None:
        pass

    def __delitem__(self, step_info: "WorkflowStepInfo") -> None:
        pass

    def __contains__(self, step_info: "WorkflowStepInfo") -> bool:
        return False


@WorkflowCache.register("memory")
class MemoryWorkflowCache(WorkflowCache):
    def __init__(self) -> None:
        self._cache: dict["WorkflowStepInfo", Any] = {}

    def __getitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]") -> T:
        return cast(T, self._cache[step_info])

    def __setitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]", value: T) -> None:
        self._cache[step_info] = value

    def __delitem__(self, step_info: "WorkflowStepInfo") -> None:
        del self._cache[step_info]

    def __contains__(self, step_info: "WorkflowStepInfo") -> bool:
        return step_info in self._cache

    def __len__(self) -> int:
        return len(self._cache)


@WorkflowCache.register("filesystem")
class FilesystemWorkflowCache(WorkflowCache):
    _LOCK_FILENAME: ClassVar[str] = "__lock__"

    def __init__(self, directory: Union[str, PathLike]) -> None:
        self._directory = Path(directory)
        self._directory.mkdir(parents=True, exist_ok=True)

    def _get_step_cache_dir(self, step_info: "WorkflowStepInfo") -> Path:
        return self._directory / step_info.fingerprint

    def _get_step_cache_lock(self, step_info: "WorkflowStepInfo") -> FileLock:
        return FileLock(str(self._get_step_cache_dir(step_info) / self._LOCK_FILENAME))

    def __getitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]") -> T:
        with self._get_step_cache_lock(step_info):
            step_cache_dir = self._get_step_cache_dir(step_info)
            return cast(T, step_info.format.read(step_cache_dir))

    def __setitem__(self, step_info: "WorkflowStepInfo[WorkflowStep[T]]", value: T) -> None:
        with self._get_step_cache_lock(step_info):
            step_cache_dir = self._get_step_cache_dir(step_info)
            step_cache_dir.mkdir(parents=True, exist_ok=True)
            step_info.format.write(value, step_cache_dir)

    def __delitem__(self, step_info: "WorkflowStepInfo") -> None:
        with self._get_step_cache_lock(step_info):
            step_cache_dir = self._get_step_cache_dir(step_info)
            for path in step_cache_dir.glob("**/*"):
                path.unlink()
            step_cache_dir.rmdir()

    def __contains__(self, step_info: "WorkflowStepInfo") -> bool:
        return self._get_step_cache_dir(step_info).exists()



from collections.abc import Iterable, Iterator
from logging import Logger
from os import PathLike
from typing import Generic, Optional, TypeVar, Union

import dill

from .processor import Processor
from .types import T_Example, T_Prediction

T_Fixtures = TypeVar("T_Fixtures")
T_SetupParams = TypeVar("T_SetupParams")
T_PredictionParams = TypeVar("T_PredictionParams")
T_EvaluationParams = TypeVar("T_EvaluationParams")
T_Model = TypeVar("T_Model", bound="Model")


class Model(
    Generic[
        T_Example,
        T_Prediction,
        T_Fixtures,
        T_SetupParams,
        T_PredictionParams,
        T_EvaluationParams,
    ],
    Processor[
        T_Example,
        T_Prediction,
        T_Fixtures,
        T_PredictionParams,
    ],
):
    Input: type[T_Example]
    Output: type[T_Prediction]
    SetupParams: type[T_SetupParams]
    PredictionParams: type[T_PredictionParams]
    EvaluationParams: type[T_EvaluationParams]

    @property
    def logger(self) -> Logger:
        from harbory.workflow import use_step_logger

        return use_step_logger(default=f"{self.__class__.__module__}.{self.__class__.__name__}")

    def setup(self, params: Optional[T_SetupParams] = None) -> None:
        pass

    def train(
        self,
        train_dataset: Iterable[T_Example],
        valid_dataset: Optional[Iterable[T_Example]],
    ) -> None:
        pass

    def predict(
        self,
        dataset: Iterable[T_Example],
        params: Optional[T_PredictionParams] = None,
        *,
        batch_size: Optional[int] = None,
        max_workers: Optional[int] = None,
    ) -> Iterator[T_Prediction]:
        yield from self(
            dataset,
            params,
            batch_size=batch_size,
            max_workers=max_workers,
        )

    def evaluate(
        self,
        dataset: Iterable[T_Example],
        params: Optional[T_EvaluationParams] = None,
    ) -> dict[str, float]:
        raise NotImplementedError

    def save(self, path: Union[str, PathLike]) -> None:
        with open(path, "wb") as f:
            dill.dump(self, f)

    @classmethod
    def load(cls: type[T_Model], path: Union[str, PathLike]) -> T_Model:
        with open(path, "rb") as f:
            model = dill.load(f)
            assert isinstance(model, cls)
            return model



import hashlib
import io
from typing import Any

import dill


def hash_object_bytes(o: Any) -> bytes:
    m = hashlib.blake2b()
    with io.BytesIO() as buffer:
        dill.dump(o, buffer)
        m.update(buffer.getbuffer())
    return m.digest()


def hash_object(o: Any) -> int:
    return int.from_bytes(hash_object_bytes(o), "big")


def murmurhash3(key: str, seed: int = 0) -> int:
    length = len(key)

    remainder = length & 3
    n = length - remainder
    h1 = seed
    c1 = 0xCC9E2D51
    c2 = 0x1B873593
    i = 0

    while i < n:
        k1 = (
            (ord(key[i]) & 0xFF)
            | ((ord(key[i + 1]) & 0xFF) << 8)
            | (ord(key[i + 2]) & 0xFF) << 16
            | (ord(key[i + 3]) & 0xFF) << 24
        )
        i += 4

        k1 = ((((k1 & 0xFFFF) * c1) + ((((k1 >> 16) * c1) & 0xFFFF) << 16))) & 0xFFFFFFFF
        k1 = (k1 << 15) | (k1 >> 17)
        k1 = ((((k1 & 0xFFFF) * c2) + ((((k1 >> 16) * c2) & 0xFFFF) << 16))) & 0xFFFFFFFF

        h1 ^= k1
        h1 = (h1 << 13) | (h1 >> 19)
        h1b = ((((h1 & 0xFFFF) * 5) + ((((h1 >> 16) * 5) & 0xFFFF) << 16))) & 0xFFFFFFFF
        h1 = ((h1b & 0xFFFF) + 0x6B64) + ((((h1b >> 16) + 0xE654) & 0xFFFF) << 16)

    k1 = 0
    if remainder == 3:
        k1 ^= (ord(key[i + 2]) & 0xFF) << 16
        k1 ^= (ord(key[i + 1]) & 0xFF) << 8
        k1 ^= ord(key[i]) & 0xFF
    elif remainder == 2:
        k1 ^= (ord(key[i + 1]) & 0xFF) << 8
        k1 ^= ord(key[i]) & 0xFF
    elif remainder == 1:
        k1 ^= ord(key[i]) & 0xFF
    k1 = ((((k1 & 0xFFFF) * c1) + ((((k1 >> 16) * c1) & 0xFFFF) << 16))) & 0xFFFFFFFF
    k1 = (k1 << 15) | (k1 >> 17)
    k1 = ((((k1 & 0xFFFF) * c2) + ((((k1 >> 16) * c2) & 0xFFFF) << 16))) & 0xFFFFFFFF
    h1 ^= k1

    h1 ^= length
    h1 ^= h1 >> 16
    h1 = ((((h1 & 0xFFFF) * 0x85EBCA6B) + ((((h1 >> 16) * 0x85EBCA6B) & 0xFFFF) << 16))) & 0xFFFFFFFF
    h1 ^= h1 >> 13
    h1 = ((((h1 & 0xFFFF) * 0xC2B2AE35) + ((((h1 >> 16) * 0xC2B2AE35) & 0xFFFF) << 16))) & 0xFFFFFFFF
    h1 ^= h1 >> 16

    return h1 >> 0



from typing import Any


def is_namedtuple(obj: Any) -> bool:
    if not isinstance(obj, type) and isinstance(obj, object):
        obj = type(obj)
    if not isinstance(obj, type):
        return False
    if not issubclass(obj, tuple):
        return False
    fields = getattr(obj, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(name) is str for name in fields)



import dataclasses
from os import PathLike
from typing import Any, ClassVar, Generic, Optional, TypeVar

from harbory.common.jsonnet import FromJsonnet

from .archive import MachineryArchive
from .colt import COLT_BUILDER
from .data import DataSink, DataSource
from .model import Model, T_EvaluationParams, T_PredictionParams, T_SetupParams
from .processor import Processor
from .types import T_Example, T_Prediction

_T_MachinerySettings = TypeVar("_T_MachinerySettings", bound="MachinerySettings")


@dataclasses.dataclass
class MachinerySettings(FromJsonnet):
    __COLT_BUILDER__: ClassVar = COLT_BUILDER


@dataclasses.dataclass
class MachineryTraningSettings(
    Generic[T_Example, T_Prediction, T_SetupParams, T_EvaluationParams],
    MachinerySettings,
):
    model: Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams]
    train_dataset: DataSource[T_Example]
    valid_dataset: Optional[DataSource[T_Example]] = None
    test_dataset: Optional[DataSource[T_Example]] = None
    setup: Optional[T_SetupParams] = None
    evaluation: Optional[T_EvaluationParams] = None
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None
    postprocessor: Optional[Processor[T_Prediction, T_Prediction, Any, Any]] = None

    def __post_init__(self) -> None:
        if self.setup is not None:
            self.setup = self.__COLT_BUILDER__(self.setup, self.model.SetupParams)
        if self.evaluation is not None:
            self.evaluation = self.__COLT_BUILDER__(self.evaluation, self.model.EvaluationParams)


@dataclasses.dataclass
class MachineryPredictionSettings(
    Generic[T_Example, T_Prediction, T_SetupParams, T_PredictionParams],
    MachinerySettings,
):
    archive: MachineryArchive[Model[T_Example, T_Prediction, Any, T_SetupParams, T_PredictionParams, Any]]
    dataset: DataSource[T_Example]
    output: DataSink[T_Prediction]
    setup: Optional[T_SetupParams] = None
    params: Optional[T_PredictionParams] = None
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None
    postprocessor: Optional[Processor[T_Prediction, T_Prediction, Any, Any]] = None
    batch_size: Optional[int] = None
    max_workers: Optional[int] = None

    @property
    def model(self) -> Model[T_Example, T_Prediction, Any, T_SetupParams, Any, Any]:
        return self.archive.model

    @classmethod
    def __pre_init__(self, config: Any) -> Any:
        if isinstance(config.get("archive"), (str, PathLike)):
            config["archive"] = MachineryArchive.load(config["archive"])
        return config

    def __post_init__(self) -> None:
        if self.setup is not None:
            self.setup = self.__COLT_BUILDER__(self.setup, self.model.SetupParams)
        if self.params is not None:
            self.params = self.__COLT_BUILDER__(self.params, self.model.PredictionParams)


@dataclasses.dataclass
class MachineryEvaluationSettings(
    Generic[T_Example, T_Prediction, T_SetupParams, T_EvaluationParams],
    MachinerySettings,
):
    archive: MachineryArchive[Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams]]
    dataset: DataSource[T_Example]
    setup: Optional[T_SetupParams] = None
    params: Optional[T_EvaluationParams] = None
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None
    batch_size: Optional[int] = None
    max_workers: Optional[int] = None

    @property
    def model(self) -> Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams]:
        return self.archive.model

    @classmethod
    def __pre_init__(self, config: Any) -> Any:
        if isinstance(config.get("archive"), (str, PathLike)):
            config["archive"] = MachineryArchive.load(config["archive"])
        return config

    def __post_init__(self) -> None:
        if self.setup is not None:
            self.setup = self.__COLT_BUILDER__(self.setup, self.model.SetupParams)
        if self.params is not None:
            self.params = self.__COLT_BUILDER__(self.params, self.model.EvaluationParams)



from collections import defaultdict
from collections.abc import Callable, Mapping
from typing import Any, Final, Generic, Optional, TypeVar, Union, cast

from colt import ColtBuilder, ColtCallback, ColtContext, ConfigurationError, Placeholder, SkipCallback
from colt.types import ParamPath
from colt.utils import get_path_name, remove_optional

from harbory.constants import COLT_ARGSKEY, COLT_TYPEKEY

from .constants import WORKFLOW_REFKEY, WORKFLOW_REFTYPE
from .step import WorkflowStep
from .types import StrictParamPath

_T = TypeVar("_T")


class WorkflowRef(Generic[_T], Placeholder[_T]):
    @staticmethod
    def is_ref(builder: "ColtBuilder", config: Any) -> bool:
        return (
            isinstance(config, Mapping)
            and set(config) == {builder.typekey, WORKFLOW_REFKEY}
            and config[builder.typekey] == WORKFLOW_REFTYPE
            and isinstance(config[WORKFLOW_REFKEY], str)
        )

    def __init__(
        self,
        annotation: _T,
        path: tuple[Union[int, str], ...],
        step_name: str,
        config: Any,
    ) -> None:
        super().__init__(annotation)
        self._path = path
        self._step_name = step_name
        self._config = config

    @property
    def path(self) -> tuple[Union[int, str], ...]:
        return self._path

    @property
    def step_name(self) -> str:
        return self._step_name

    @property
    def config(self) -> Any:
        return self._config


class RefCallback(ColtCallback):
    """
    Replace `ref` configs with `WorkflowRef` instances as placeholders
    """

    _STEP_NAME_TO_TYPE_KEY: Final[str] = "__harbory_workflow_step__"

    def _get_step_name_to_type(self, graph_path: ParamPath, context: ColtContext) -> dict[str, type[WorkflowStep[Any]]]:
        registry: dict[ParamPath, dict[str, type[WorkflowStep[Any]]]] = context.state.setdefault(
            self._STEP_NAME_TO_TYPE_KEY, defaultdict(dict)
        )
        return registry[graph_path]

    def _register_step_types(
        self,
        builder: "ColtBuilder",
        graph_path: StrictParamPath,
        config: Mapping,
        context: ColtContext,
    ) -> None:
        steps = config.get("steps")
        step_path = graph_path + ("steps",)
        if steps is None:
            raise ConfigurationError(f"[{get_path_name(graph_path)}] Missing 'steps' key")
        if not isinstance(steps, Mapping):
            raise ConfigurationError(f"[{get_path_name(step_path)}] Expected a mapping, got {steps}")
        step_name_to_type = self._get_step_name_to_type(graph_path, context)
        for step_name, step_config in steps.items():
            if not isinstance(step_name, str):
                raise ConfigurationError(f"[{get_path_name(step_path)}] Expected a string key, got {step_name}")
            step_config_path = graph_path + ("steps", step_name)
            if not isinstance(step_config, Mapping):
                raise ConfigurationError(f"[{get_path_name(step_config_path)}] Expected a mapping")
            step_type = step_config.get(builder.typekey)
            step_type_path = step_config_path + (builder.typekey,)
            if step_type is None:
                raise ConfigurationError(f"[{get_path_name(step_config_path)}] Missing '{builder.typekey}' key")
            if not isinstance(step_type, str):
                raise ConfigurationError(f"[{get_path_name(step_type_path)}] Expected a string, got {step_type}")
            step = cast(type[WorkflowStep[Any]], WorkflowStep[Any].by_name(step_type))
            step_name_to_type[step_name] = step

    def _find_step_type(
        self, path: tuple[Union[int, str], ...], step_name: str, context: ColtContext
    ) -> type[WorkflowStep[Any]]:
        registry: dict[ParamPath, dict[str, type[WorkflowStep[Any]]]] = context.state[self._STEP_NAME_TO_TYPE_KEY]
        graph_path = path
        while graph_path:
            if graph_path in registry:
                break
            graph_path = graph_path[:-1]
        if graph_path not in registry:
            raise ConfigurationError(f"[get_path_name(path)] Step '{step_name}' not found")
        step_name_to_type = registry[graph_path]
        if step_name not in step_name_to_type:
            raise ConfigurationError(f"[get_path_name(path)] Step '{step_name}' not found")
        return step_name_to_type[step_name]

    def on_build(
        self,
        path: ParamPath,
        config: Any,
        builder: ColtBuilder,
        context: ColtContext,
        annotation: Optional[Union[type[_T], Callable[..., _T]]] = None,
    ) -> Any:
        from .graph import WorkflowGraph

        annotation = remove_optional(annotation)
        if config and isinstance(annotation, type) and issubclass(annotation, WorkflowGraph):
            if not isinstance(config, Mapping):
                raise ConfigurationError(f"[{get_path_name(path)}] Expected a mapping, got {config}")
            self._register_step_types(builder, path, config, context)
            return config

        if WorkflowRef.is_ref(builder, config):
            step_name = config[WORKFLOW_REFKEY]
            step_type = self._find_step_type(path, step_name, context)
            return WorkflowRef(step_type.get_output_type(), path, step_name, config)

        raise SkipCallback


COLT_BUILDER: Final = ColtBuilder(
    typekey=COLT_TYPEKEY,
    argskey=COLT_ARGSKEY,
    callback=RefCallback(),
)



import contextvars
import shutil
from collections.abc import Sequence
from io import StringIO
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import ClassVar, Optional, TypeVar, Union, cast

import mlflow
from filelock import FileLock
from mlflow.entities import Experiment as MlflowExperiment
from mlflow.tracking.client import MlflowClient

from harbory.common.logutils import LogCapture
from harbory.workflow import (
    WorkflowCache,
    WorkflowCallback,
    WorkflowExecutionContext,
    WorkflowExecutionID,
    WorkflowExecutionInfo,
    WorkflowExecutor,
    WorkflowGraph,
    WorkflowOrganizer,
    WorkflowStep,
    WorkflowStepContext,
    WorkflowStepInfo,
    get_step_logger_from_info,
    use_step_context,
)
from harbory.workflow.colt import COLT_BUILDER
from harbory.workflow.utils import as_jsonvalue

from . import utils as mlflow_utils
from .constants import DEFAULT_MLFLOW_DIRECTORY, DEFAULT_MLFLOW_EXPERIMENT_NAME
from .utils import MlflowRun, MlflowRunStatus, MlflowTag, WorkflowCacheStatus, fetch_mlflow_run, get_mlflow_experiment

logger = getLogger(__name__)

T = TypeVar("T")

_MLFLOW_EXPERIMENT = contextvars.ContextVar[Optional[MlflowExperiment]]("_MLFLOW_EXPERIMENT", default=None)


@WorkflowCache.register("mlflow")
class MlflowWorkflowCache(WorkflowCache):
    _LOCK_FILENAME: ClassVar[str] = "__lock__"
    _CACHE_DIRNAME: ClassVar[str] = "cache"
    _DEFAULT_DIRECTORY: ClassVar[Path] = DEFAULT_MLFLOW_DIRECTORY / "cache"

    def __init__(
        self,
        experiment_name: str = DEFAULT_MLFLOW_EXPERIMENT_NAME,
        directory: Optional[Union[str, PathLike]] = None,
        mlflow_client: Optional[MlflowClient] = None,
    ) -> None:
        self._client = mlflow_client or MlflowClient()
        self._experiment_name = experiment_name
        self._directory = Path(directory or self._DEFAULT_DIRECTORY)
        self._directory.mkdir(parents=True, exist_ok=True)

    def _get_step_cache_dir(self, step_info: WorkflowStepInfo) -> Path:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            step_info=step_info,
        )
        mlflow_artifact_dir = mlflow_utils.get_mlflow_local_artifact_storage_path(run) if run is not None else None
        if mlflow_artifact_dir is not None:
            return mlflow_artifact_dir / self._CACHE_DIRNAME
        return self._directory / step_info.fingerprint

    def _get_step_cache_lock(self, step_info: WorkflowStepInfo) -> FileLock:
        return FileLock(str(self._get_step_cache_dir(step_info) / self._LOCK_FILENAME))

    def __getitem__(self, step_info: WorkflowStepInfo[WorkflowStep[T]]) -> T:
        if step_info not in self:
            raise KeyError(step_info)
        step_cache_dir = self._get_step_cache_dir(step_info)
        with self._get_step_cache_lock(step_info):
            if not step_cache_dir.exists():
                try:
                    mlflow_utils.download_mlflow_artifacts(
                        self._client,
                        self._experiment_name,
                        step_info,
                        step_cache_dir,
                        artifact_path=self._CACHE_DIRNAME,
                    )
                except FileNotFoundError:
                    raise KeyError(step_info)
            return cast(T, step_info.format.read(step_cache_dir))

    def __setitem__(self, step_info: WorkflowStepInfo[WorkflowStep[T]], value: T) -> None:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            step_info=step_info,
        )
        if run is None:
            raise RuntimeError(f"Run for step {step_info} not found")
        elif run.info.status != MlflowRunStatus.RUNNING.value:
            raise ValueError(f"Run {run.info.run_id} is not running")
        step_cache_dir = self._get_step_cache_dir(step_info)
        with self._get_step_cache_lock(step_info):
            mlflow_utils.update_mlflow_tags(
                self._client,
                run,
                {MlflowTag.MLFACTORY_STEP_CACHE_STATUS: WorkflowCacheStatus.PENDING.value},
            )
            try:
                step_cache_dir.mkdir(parents=True, exist_ok=True)
                step_info.format.write(value, step_cache_dir)
                if not mlflow_utils.is_mlflow_using_local_artifact_storage(run):
                    mlflow_utils.upload_mlflow_artifacts(
                        self._client,
                        self._experiment_name,
                        step_info,
                        step_cache_dir,
                        artifact_path=self._CACHE_DIRNAME,
                    )
            except Exception as e:
                mlflow_utils.update_mlflow_tags(
                    self._client,
                    run,
                    {MlflowTag.MLFACTORY_STEP_CACHE_STATUS: WorkflowCacheStatus.INACTIVE.value},
                )
                raise e
            else:
                mlflow_utils.update_mlflow_tags(
                    self._client,
                    run,
                    {MlflowTag.MLFACTORY_STEP_CACHE_STATUS: WorkflowCacheStatus.ACTIVE.value},
                )

    def __delitem__(self, step_info: WorkflowStepInfo) -> None:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            step_info=step_info,
        )
        if run is None:
            raise KeyError(step_info)
        mlflow_utils.update_mlflow_tags(
            self._client,
            run,
            {MlflowTag.MLFACTORY_STEP_CACHE_STATUS: WorkflowCacheStatus.INACTIVE.value},
        )
        if not mlflow_utils.is_mlflow_using_local_artifact_storage(run):
            step_cache_dir = self._get_step_cache_dir(step_info)
            with self._get_step_cache_lock(step_info):
                shutil.rmtree(step_cache_dir, ignore_errors=True)

    def __contains__(self, step_info: WorkflowStepInfo) -> bool:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            step_info=step_info,
        )
        return (
            run is not None
            and run.data.tags.get(MlflowTag.MLFACTORY_STEP_CACHE_STATUS) == WorkflowCacheStatus.ACTIVE.value
        )


@WorkflowCallback.register("mlflow")
class MlflowWorkflowCallback(WorkflowCallback):
    _LOG_FILENAME: ClassVar[str] = "out.log"
    _STEP_METADATA_ARTIFACT_FILENAME: ClassVar[str] = "step.json"
    _EXECUTION_METADATA_ARTIFACT_FILENAME: ClassVar[str] = "execution.json"

    def __init__(
        self,
        experiment_name: str = DEFAULT_MLFLOW_EXPERIMENT_NAME,
        mlflow_client: Optional[MlflowClient] = None,
    ) -> None:
        self._client = mlflow_client or MlflowClient()
        self._experiment_name = experiment_name
        self._execution_run: Optional[MlflowRun] = None
        self._execution_log: Optional[LogCapture[StringIO]] = None
        self._step_log: dict[WorkflowStepInfo, LogCapture[StringIO]] = {}

    def on_execution_start(
        self,
        execution_context: WorkflowExecutionContext,
    ) -> None:
        assert self._execution_run is None
        execution_info = execution_context.info
        if execution_info.id is None:
            execution_info.id = mlflow_utils.generate_new_execution_id(self._client, self._experiment_name)
        self._execution_log = LogCapture(StringIO())
        self._execution_log.start()
        self._execution_run = mlflow_utils.add_mlflow_run(
            self._client,
            self._experiment_name,
            execution_info,
        )
        self._client.log_dict(
            run_id=self._execution_run.info.run_id,
            dictionary=cast(dict, as_jsonvalue(execution_info)),
            artifact_file=self._EXECUTION_METADATA_ARTIFACT_FILENAME,
        )

    def on_execution_end(
        self,
        execution_context: "WorkflowExecutionContext",
    ) -> None:
        assert self._execution_run is not None
        mlflow_utils.terminate_mlflow_run(
            self._client,
            self._experiment_name,
            execution_context.state,
        )
        if self._execution_log is not None:
            self._execution_log.stop()
            self._client.log_text(
                run_id=self._execution_run.info.run_id,
                text=self._execution_log.stream.getvalue(),
                artifact_file=self._LOG_FILENAME,
            )
            self._execution_log.stream.close()
        self._execution_run = None

    def on_step_start(
        self,
        step_context: WorkflowStepContext,
        execution_context: WorkflowExecutionContext,
    ) -> None:
        assert self._execution_run is not None
        step_info = step_context.info
        run = mlflow_utils.add_mlflow_run(
            self._client,
            self._experiment_name,
            step_info,
            parent_run_id=self._execution_run.info.run_id,
        )
        self._client.log_dict(
            run_id=run.info.run_id,
            dictionary=cast(dict, as_jsonvalue(step_info)),
            artifact_file=self._STEP_METADATA_ARTIFACT_FILENAME,
        )
        self._step_log[step_info] = LogCapture(StringIO(), logger=get_step_logger_from_info(step_info))
        self._step_log[step_info].start()

    def on_step_end(
        self,
        step_context: WorkflowStepContext,
        execution_context: WorkflowExecutionContext,
    ) -> None:
        step_info = step_context.info
        mlflow_utils.terminate_mlflow_run(
            self._client,
            self._experiment_name,
            step_context.state,
        )
        if (step_log := self._step_log.pop(step_info, None)) is not None:
            run = mlflow_utils.fetch_mlflow_run(
                self._client,
                self._experiment_name,
                step_info=step_info,
            )
            if run is None:
                raise RuntimeError(f"Run for step {step_info} not found")
            step_log.stop()
            self._client.log_text(
                run_id=run.info.run_id,
                text=step_log.stream.getvalue(),
                artifact_file=self._LOG_FILENAME,
            )
            step_log.stream.close()


@WorkflowOrganizer.register("mlflow")
class MlflowWorkflowOrganizer(WorkflowOrganizer):
    def __init__(
        self,
        experiment_name: str = DEFAULT_MLFLOW_EXPERIMENT_NAME,
        cache: Optional[WorkflowCache] = None,
        callbacks: Optional[Union[WorkflowCallback, Sequence[WorkflowCallback]]] = None,
    ) -> None:
        self._client = MlflowClient()
        self._experiment_name = experiment_name

        cache = cache or MlflowWorkflowCache(
            experiment_name=experiment_name,
            mlflow_client=self._client,
        )
        if callbacks is None:
            callbacks = []
        elif isinstance(callbacks, WorkflowCallback):
            callbacks = [callbacks]
        if not any(isinstance(callback, MlflowWorkflowCallback) for callback in callbacks):
            mlflow_callback = MlflowWorkflowCallback(
                experiment_name,
                mlflow_client=self._client,
            )
            callbacks = [mlflow_callback] + list(callbacks)

        super().__init__(cache, callbacks)

    def run(
        self,
        executor: WorkflowExecutor,
        execution: Union[WorkflowGraph, WorkflowExecutionInfo],
    ) -> WorkflowExecutionInfo:
        cxt = contextvars.copy_context()

        super_run = super().run

        def _run() -> WorkflowExecutionInfo:
            experiment = get_mlflow_experiment(self._experiment_name)
            _MLFLOW_EXPERIMENT.set(experiment)
            return super_run(executor, execution)

        return cxt.run(_run)

    def get(self, execution_id: WorkflowExecutionID) -> Optional[WorkflowExecutionInfo]:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            execution_info=execution_id,
        )
        if run is None:
            return None
        artifact_uri = run.info.artifact_uri
        execution_dict = mlflow.artifacts.load_dict(
            artifact_uri + "/" + MlflowWorkflowCallback._EXECUTION_METADATA_ARTIFACT_FILENAME
        )
        execution_info = COLT_BUILDER(execution_dict, WorkflowExecutionInfo)
        return execution_info

    def exists(self, execution_id: WorkflowExecutionID) -> bool:
        run = mlflow_utils.fetch_mlflow_run(
            self._client,
            self._experiment_name,
            execution_info=execution_id,
        )
        return run is not None

    def remove(self, execution_id: WorkflowExecutionID) -> None:
        for run in mlflow_utils.fetch_mlflow_runs(
            self._client,
            self._experiment_name,
            execution_info=execution_id,
            with_children=True,
        ):
            logger.info(f"Removing run {run.info.run_id}")
            self._client.delete_run(run.info.run_id)


class MlflowLogger:
    def __init__(self, run: MlflowRun):
        self.run = run

    @property
    def mlflow_client(self) -> MlflowClient:
        return MlflowClient()

    def log_metric(
        self,
        key: str,
        value: float,
        timestamp: Optional[int] = None,
        step: Optional[int] = None,
    ) -> None:
        self.mlflow_client.log_metric(
            run_id=self.run.info.run_id,
            key=key,
            value=value,
            timestamp=timestamp,
            step=step,
        )

    def log_metrics(self, metrics: dict[str, float]) -> None:
        for key, value in metrics.items():
            self.log_metric(key, value)


def use_mlflow_experiment() -> Optional[MlflowExperiment]:
    return _MLFLOW_EXPERIMENT.get()


def use_mlflow_logger() -> Optional[MlflowLogger]:
    if (experiment := use_mlflow_experiment()) is None:
        return None

    if (context := use_step_context()) is None:
        return None

    client = MlflowClient()
    if (run := fetch_mlflow_run(client, experiment, step_info=context.info)) is None:
        return None

    return MlflowLogger(run)



import itertools
import math
from collections import abc
from collections.abc import Callable, Iterable, Iterator
from typing import Any, Generic, TypeVar

T = TypeVar("T")


class SizedIterator(Generic[T]):
    """
    A wrapper for an iterator that knows its size.

    Args:
        iterator: The iterator.
        size: The size of the iterator.
    """

    def __init__(self, iterator: Iterator[T], size: int):
        self.iterator = iterator
        self.size = size

    def __iter__(self) -> Iterator[T]:
        return self.iterator

    def __next__(self) -> T:
        return next(self.iterator)

    def __len__(self) -> int:
        return self.size


def batched(iterable: Iterable[T], batch_size: int, drop_last: bool = False) -> Iterator[list[T]]:
    """
    Batch an iterable into lists of the given size.

    Args:
        iterable: The iterable.
        batch_size: The size of each batch.
        drop_last: Whether to drop the last batch if it is smaller than the given size.

    Returns:
        An iterator over batches.
    """

    def iterator() -> Iterator[list[T]]:
        batch = []
        for item in iterable:
            batch.append(item)
            if len(batch) == batch_size:
                yield batch
                batch = []
        if batch and not drop_last:
            yield batch

    if isinstance(iterable, abc.Sized):
        num_batches = math.ceil(len(iterable) / batch_size)
        return SizedIterator(iterator(), num_batches)

    return iterator()


def batched_iterator(iterable: Iterable[T], batch_size: int) -> Iterator[Iterator[T]]:
    """
    Batch an iterable into iterators of the given size.

    Args:
        iterable: The iterable.
        batch_size: The size of each batch.

    Returns:
        An iterator over batches.
    """

    def iterator() -> Iterator[Iterator[T]]:
        stop = False
        batch_progress = 0

        def iterator_wrapper() -> Iterator[T]:
            nonlocal batch_progress
            for item in iterable:
                yield item
                batch_progress += 1

        iterator = iterator_wrapper()

        def consume(n: int) -> Iterator[T]:
            for _ in range(n):
                try:
                    yield next(iterator)
                except StopIteration:
                    nonlocal stop
                    stop = True
                    break

        while not stop:
            try:
                batch_progress = 0
                yield itertools.chain([next(iterator)], consume(batch_size - 1))
                for _ in range(batch_size - batch_progress):
                    next(iterator)
            except StopIteration:
                break

    if isinstance(iterable, abc.Sized):
        num_batches = math.ceil(len(iterable) / batch_size)
        return SizedIterator(iterator(), num_batches)

    return iterator()


def iter_with_callback(
    iterable: Iterable[T],
    callback: Callable[[T], Any],
) -> Iterator[T]:
    """
    Iterate over an iterable and call a callback for each item.

    Args:
        iterable: The iterable.
        callback: The callback to call for each item.

    Returns:
        An iterator over the iterable.
    """

    def iterator() -> Iterator[T]:
        for item in iterable:
            yield item
            callback(item)

    if isinstance(iterable, abc.Sized):
        return SizedIterator(iterator(), len(iterable))

    return iterator()


def wrap_iterator(wrapper: Callable[[Iterable[T]], Iterator[T]], iterable: Iterable[T]) -> Iterator:
    """
    Wrap an iterator with a function.

    Note:
        This function assume that the wrapped iterator is of the same size as the input iterator.

    Args:
        wrapper: The function to wrap the iterator.

    Returns:
        An iterator wrapped with the function.
    """

    def wrapped() -> Iterator[T]:
        return wrapper(iterable)

    if isinstance(iterable, abc.Sized):
        return SizedIterator(wrapped(), len(iterable))

    return wrapped()



import sys
from collections.abc import Iterator, Mapping
from typing import Any, TextIO, TypedDict

from colt import ConfigurationError, Lazy

from harbory.common.dag import DAG
from harbory.common.jsonnet import FromJsonnet
from harbory.types import JsonValue

from .colt import COLT_BUILDER, WorkflowRef
from .constants import WORKFLOW_REFKEY
from .step import WorkflowStep, WorkflowStepInfo
from .types import StrictParamPath


class WorkflowGraphConfig(TypedDict):
    steps: dict[str, JsonValue]


class WorkflowGraph(FromJsonnet):
    __COLT_BUILDER__ = COLT_BUILDER

    @classmethod
    def _build_step_info(
        cls,
        steps: Mapping[str, Lazy[WorkflowStep]],
    ) -> Mapping[str, WorkflowStepInfo]:
        if not steps:
            return {}

        builder = next(iter(steps.values()))._builder

        def find_dependencies(obj: Any, path: tuple[str, ...]) -> frozenset[tuple[StrictParamPath, str]]:
            refs: set[tuple[StrictParamPath, str]] = set()
            if WorkflowRef.is_ref(builder, obj):
                step_name = str(obj[WORKFLOW_REFKEY])
                refs |= {(path, step_name)}
            if isinstance(obj, Mapping):
                for key, value in obj.items():
                    refs |= find_dependencies(value, path + (key,))
            if isinstance(obj, (list, tuple)):
                for i, value in enumerate(obj):
                    refs |= find_dependencies(value, path + (str(i),))
            return frozenset(refs)

        dependencies = {name: find_dependencies(lazy_step.config, ()) for name, lazy_step in steps.items()}

        stack: set[str] = set()
        visited: set[str] = set()
        sorted_step_names: list[str] = []

        def topological_sort(name: str) -> None:
            if name in stack:
                raise ConfigurationError(f"Cycle detected in workflow dependencies: {name} -> {stack}")
            if name in visited:
                return
            stack.add(name)
            visited.add(name)
            for _, dep_name in dependencies[name]:
                topological_sort(dep_name)
            stack.remove(name)
            sorted_step_names.append(name)

        for name in steps.keys():
            topological_sort(name)

        step_name_to_info: dict[str, WorkflowStepInfo] = {}
        for name in sorted_step_names:
            step = steps[name]
            step_dependencies = frozenset((path, step_name_to_info[dep_name]) for path, dep_name in dependencies[name])
            step_name_to_info[name] = WorkflowStepInfo(name, step, step_dependencies)

        return step_name_to_info

    def __init__(
        self,
        steps: Mapping[str, Lazy[WorkflowStep]],
    ) -> None:
        self._step_info = self._build_step_info(steps)

    def __iter__(self) -> Iterator[WorkflowStepInfo]:
        return iter(self._step_info.values())

    def __getitem__(self, step_name: str) -> WorkflowStepInfo:
        return self._step_info[step_name]

    def get_subgraph(self, step_name: str) -> "WorkflowGraph":
        if step_name not in self._step_info:
            raise ValueError(f"Step {step_name} not found in the graph")
        step_info = self._step_info[step_name]
        subgraph_steps: dict[str, Lazy[WorkflowStep]] = {step_name: step_info.step}
        for _, dependant_step_info in step_info.dependencies:
            for sub_step_info in self.get_subgraph(dependant_step_info.name):
                subgraph_steps[sub_step_info.name] = sub_step_info.step
        return WorkflowGraph(subgraph_steps)

    def visualize(
        self,
        *,
        output: TextIO = sys.stdout,
    ) -> None:
        dag = DAG({name: {dep.name for _, dep in info.dependencies} for name, info in self._step_info.items()})
        dag.visualize(output=output)

    def to_dict(self) -> dict[str, Any]:
        return {"steps": {step_info.name: step_info.step.config for step_info in self}}

    @classmethod
    def from_config(self, config: WorkflowGraphConfig) -> "WorkflowGraph":
        return self.__COLT_BUILDER__(config, WorkflowGraph)



import logging
import os

from harbory.commands import main

if os.environ.get("HARBORY_DEBUG"):
    LEVEL = logging.DEBUG
else:
    level_name = os.environ.get("HARBORY_LOG_LEVEL", "INFO")
    LEVEL = logging._nameToLevel.get(level_name, logging.INFO)

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(name)s - %(message)s", level=LEVEL)


def run() -> None:
    main(prog="harbory")


if __name__ == "__main__":
    run()



from .archive import MachineryArchive  # noqa: F401
from .data import DataSink, DataSource  # noqa: F401
from .format import MachineryArchiveFormat  # noqa: F401
from .model import Model  # noqa: F401
from .processor import (  # noqa: F401
    CallableProcessor,
    ChainProcessor,
    ComposeProcessor,
    FieldProcessor,
    PassThroughProcessor,
    Processor,
)
from .settings import (  # noqa: F401
    MachineryEvaluationSettings,
    MachineryPredictionSettings,
    MachinerySettings,
    MachineryTraningSettings,
)
from .step import evaluate, predict, train  # noqa: F401



from pathlib import Path
from typing import Any, ClassVar, Generic, cast

from harbory.workflow import Format

from .archive import MachineryArchive, T_MachineryArchive


@Format.register("harbory.machinery.archive")
class MachineryArchiveFormat(Format[T_MachineryArchive], Generic[T_MachineryArchive]):
    _ARTIFACT_FILENAME: ClassVar[str] = "artifact.tar.gz"

    def write(self, artifact: T_MachineryArchive, directory: Path) -> None:
        artifact.save(directory / self._ARTIFACT_FILENAME)

    def read(self, directory: Path) -> T_MachineryArchive:
        return cast(T_MachineryArchive, MachineryArchive.load(directory / self._ARTIFACT_FILENAME))

    @classmethod
    def is_default_of(cls, obj: Any) -> bool:
        return isinstance(obj, MachineryArchive)



import dataclasses
from typing import Any, ClassVar, Protocol, TypeVar, Union, runtime_checkable

from pydantic import BaseModel, JsonValue  # noqa: F401

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)

DataContainer = Union["IDataclass", "INamedTuple", BaseModel, dict[str, Any]]

S_DataContainer = TypeVar("S_DataContainer", bound=DataContainer)
T_DataContainer = TypeVar("T_DataContainer", bound=DataContainer)

T_NamedTuple = TypeVar("T_NamedTuple", bound="INamedTuple")


@runtime_checkable
class IDataclass(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field]]


@runtime_checkable
class INamedTuple(Protocol):
    _fields: ClassVar[tuple[str, ...]]

    def _asdict(self) -> dict[str, Any]: ...

    def _replace(self: T_NamedTuple, **kwargs: Any) -> T_NamedTuple: ...



import dataclasses
import os
from logging import getLogger
from os import PathLike
from typing import ClassVar, Mapping, Optional, Sequence, TypeVar, Union

import yaml
from colt import ColtBuilder, import_modules

from harbory.constants import DEFAULT_MLFACTORY_SETTINGS_PATH
from harbory.workflow import WorkflowSettings

from .constants import COLT_ARGSKEY, COLT_TYPEKEY

logger = getLogger(__name__)

T_HarborySettings = TypeVar("T_HarborySettings", bound="HarborySettings")


@dataclasses.dataclass(frozen=True)
class HarborySettings:
    __COLT_BUILDER__: ClassVar[ColtBuilder] = ColtBuilder(typekey=COLT_TYPEKEY, argskey=COLT_ARGSKEY)

    workflow: WorkflowSettings = dataclasses.field(default_factory=WorkflowSettings)

    environment: Mapping[str, str] = dataclasses.field(default_factory=dict)
    required_modules: Sequence[str] = dataclasses.field(default_factory=list)

    @classmethod
    def from_file(cls: type[T_HarborySettings], path: Union[str, PathLike]) -> T_HarborySettings:
        with open(path, "r") as f:
            settings = yaml.safe_load(f)
        # load required modules
        required_modules = cls.__COLT_BUILDER__(settings.pop("required_modules", []), Sequence[str])
        import_modules(required_modules)
        print("imported:", required_modules)
        # load environment variables
        environment = cls.__COLT_BUILDER__(settings.pop("environment", {}), Mapping[str, str])
        os.environ.update(environment)
        return cls.__COLT_BUILDER__(settings, cls)


def load_harbory_settings(path: Optional[Union[str, PathLike]] = None) -> HarborySettings:
    if path is not None or DEFAULT_MLFACTORY_SETTINGS_PATH.exists():
        path = path or DEFAULT_MLFACTORY_SETTINGS_PATH
        logger.info(f"Load harbory settings from {path}")
        return HarborySettings.from_file(path)
    return HarborySettings()



import dataclasses
import importlib


@dataclasses.dataclass(frozen=True)
class PackageInfo:
    name: str
    version: str


def get_installed_packages() -> list[PackageInfo]:
    distributions = importlib.metadata.distributions()
    return sorted([PackageInfo(d.metadata["Name"], d.version) for d in distributions], key=lambda p: p.name)



import argparse
import json
from logging import getLogger
from pathlib import Path

from harbory.constants import DEFAULT_MLFACTORY_SETTINGS_PATH
from harbory.machinery import (
    MachineryEvaluationSettings,
    MachineryPredictionSettings,
    MachineryTraningSettings,
    evaluate,
    predict,
    train,
)
from harbory.settings import HarborySettings

from .subcommand import Subcommand

logger = getLogger(__name__)


@Subcommand.register("machinery")
class MachineryCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "--settings",
            type=Path,
            default=DEFAULT_MLFACTORY_SETTINGS_PATH,
            help="settings file path",
        )


@MachineryCommand.register("train")
class MachineryTrainCommand(MachineryCommand):
    def setup(self) -> None:
        self.parser.add_argument("config", type=str, help="config file path")
        self.parser.add_argument("archive", type=str, help="archive file path")
        self.parser.add_argument("--overrides", type=str, default=None, help="overrides jsonnet file path")

    def run(self, args: argparse.Namespace) -> None:
        logger.info(f"Loading harbory settings from {args.settings}...")
        _ = HarborySettings.from_file(args.settings)

        logger.info(f"Loading settings from {args.config}...")
        settings = MachineryTraningSettings.from_jsonnet(  # type: ignore[var-annotated]
            args.config, overrides=args.overrides
        )

        logger.info("Start training...")
        archive = train(
            model=settings.model,
            train_dataset=settings.train_dataset,
            valid_dataset=settings.valid_dataset,
            test_dataset=settings.test_dataset,
            setup=settings.setup,
            evaluation=settings.evaluation,
            preprocessor=settings.preprocessor,
        )

        logger.info(f"Save archive to {args.archive}...")
        archive.save(args.archive)

        logger.info("Training completed.")


@MachineryCommand.register("predict")
class MachineryPredictCommand(MachineryCommand):
    def setup(self) -> None:
        self.parser.add_argument("config", type=str, help="config file path")
        # self.parser.add_argument("archive", type=str, help="archive file path")
        self.parser.add_argument("--overrides", type=str, default=None, help="overrides jsonnet file path")

    def run(self, args: argparse.Namespace) -> None:
        logger.info(f"Loading harbory settings from {args.settings}...")
        _ = HarborySettings.from_file(args.settings)

        logger.info(f"Loading training settings from {args.config}...")
        settings = MachineryPredictionSettings.from_jsonnet(  # type: ignore[var-annotated]
            args.config,
            overrides=args.overrides,
        )

        predictions = predict(
            model=settings.model,
            dataset=settings.dataset,
            setup=settings.setup,
            params=settings.params,
            preprocessor=settings.preprocessor,
            postprocessor=settings.postprocessor,
            batch_size=settings.batch_size,
            max_workers=settings.max_workers,
        )

        settings.output.save(predictions)


@MachineryCommand.register("evaluate")
class MachineryEvaluateCommand(MachineryCommand):
    def setup(self) -> None:
        self.parser.add_argument("config", type=str, help="config file path")
        self.parser.add_argument("--overrides", type=str, default=None, help="overrides jsonnet file path")

    def run(self, args: argparse.Namespace) -> None:
        logger.info(f"Loading harbory settings from {args.settings}...")
        _ = HarborySettings.from_file(args.settings)

        logger.info(f"Loading training settings from {args.config}...")
        settings = MachineryEvaluationSettings.from_jsonnet(  # type: ignore[var-annotated]
            args.config,
            overrides=args.overrides,
        )

        metrics = evaluate(
            model=settings.model,
            dataset=settings.dataset,
            setup=settings.setup,
            preprocessor=settings.preprocessor,
        )

        print(json.dumps(metrics, indent=2, ensure_ascii=False))



import argparse
from typing import Optional

from harbory import __version__
from harbory.commands import machinery  # noqa: F401
from harbory.commands import workflow  # noqa: F401
from harbory.commands.subcommand import Subcommand


def create_subcommand(prog: Optional[str] = None) -> Subcommand:
    parser = argparse.ArgumentParser(usage="%(prog)s", prog=prog)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
    )
    return Subcommand(parser)


def main(prog: Optional[str] = None) -> None:
    app = create_subcommand(prog)
    app()



from collections.abc import Iterator
from typing import Any, Optional, Union

from harbory.workflow import step, use_step_logger

from .archive import MachineryArchive
from .data import DataSource
from .format import MachineryArchiveFormat
from .model import Model, T_EvaluationParams, T_PredictionParams, T_SetupParams
from .processor import Processor
from .types import T_Example, T_Prediction


@step(
    "harbory.machinery.train",
    format=MachineryArchiveFormat(),
)
def train(
    model: Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams],
    train_dataset: DataSource[T_Example],
    valid_dataset: Optional[DataSource[T_Example]] = None,
    test_dataset: Optional[DataSource[T_Example]] = None,
    setup: Optional[T_SetupParams] = None,
    evaluation: Optional[T_EvaluationParams] = None,
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None,
) -> MachineryArchive[Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams]]:
    logger = use_step_logger(default="harbory.machinery.train")

    logger.info("Loading datasets...")
    train_dataset_iterator = train_dataset.load()
    valid_dataset_iterator = valid_dataset.load() if valid_dataset is not None else None
    test_dataset_iterator = test_dataset.load() if test_dataset is not None else None

    if preprocessor is not None:
        train_dataset_iterator = preprocessor(train_dataset_iterator)
        if valid_dataset_iterator is not None:
            valid_dataset_iterator = preprocessor(valid_dataset_iterator)
        if test_dataset_iterator is not None:
            test_dataset_iterator = preprocessor(test_dataset_iterator)

    logger.info("Start training...")
    model.setup(setup)
    model.train(train_dataset_iterator, valid_dataset_iterator)

    if test_dataset_iterator is not None:
        logger.info("Start evaluation on test dataset...")
        metrics = model.evaluate(test_dataset_iterator)
        logger.info(f"Test metrics: {metrics}")

    return MachineryArchive(model=model)


@step("harbory.machinery.predict")
def predict(
    model: Union[
        Model[T_Example, T_Prediction, Any, Any, T_PredictionParams, Any],
        MachineryArchive[Model[T_Example, T_Prediction, Any, Any, T_PredictionParams, Any]],
    ],
    dataset: DataSource[T_Example],
    setup: Optional[T_PredictionParams] = None,
    params: Optional[T_PredictionParams] = None,
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None,
    postprocessor: Optional[Processor[T_Prediction, T_Prediction, Any, Any]] = None,
    batch_size: Optional[int] = None,
    max_workers: Optional[int] = None,
) -> Iterator[T_Prediction]:
    logger = use_step_logger(default="harbory.machinery.predict")

    if isinstance(model, MachineryArchive):
        model = model.model

    model.setup(setup)

    logger.info("Loading dataset...")
    dataset_iterator = dataset.load()

    if preprocessor is not None:
        dataset_iterator = preprocessor(dataset_iterator)

    logger.info("Start prediction...")
    prediction_iterator = model.predict(
        dataset_iterator,
        params=params,
        batch_size=batch_size,
        max_workers=max_workers,
    )

    if postprocessor is not None:
        prediction_iterator = postprocessor(prediction_iterator)

    return prediction_iterator


@step("harbory.machinery.evaluate")
def evaluate(
    model: Union[
        Model[T_Example, T_Prediction, Any, T_SetupParams, Any, T_EvaluationParams],
        MachineryArchive[Model[T_Example, T_Prediction, Any, Any, Any, T_EvaluationParams]],
    ],
    dataset: DataSource[T_Example],
    setup: Optional[T_SetupParams] = None,
    params: Optional[T_EvaluationParams] = None,
    preprocessor: Optional[Processor[T_Example, T_Example, Any, Any]] = None,
) -> dict[str, Any]:
    logger = use_step_logger(default="harbory.machinery.evaluate")

    if isinstance(model, MachineryArchive):
        model = model.model

    model.setup(setup)

    logger.info("Loading dataset...")
    dataset_iterator = dataset.load()

    if preprocessor is not None:
        dataset_iterator = preprocessor(dataset_iterator)

    logger.info("Start evaluation...")
    metrics = model.evaluate(dataset_iterator, params=params)

    return metrics
